# -*- coding: utf-8 -*-
from unittest.mock import call, Mock

import pytest
from unittest.mock import sentinel, MagicMock

from sceptre.resolvers import (
    Resolver,
    ResolvableContainerProperty,
    ResolvableValueProperty,
    RecursiveResolve
)
from sceptre.resolvers.placeholders import use_resolver_placeholders_on_error, create_placeholder_value, PlaceholderType


class MockResolver(Resolver):
    """
    MockResolver inherits from the abstract base class Resolver, and
    implements the abstract methods. It is used to allow testing on
    Resolver, which is not otherwise instantiable.
    """

    def resolve(self):
        pass


class MockClass(object):
    resolvable_container_property = ResolvableContainerProperty("resolvable_container_property")
    container_with_alphanum_placeholder = ResolvableContainerProperty(
        "container_with_placeholder_override",
        PlaceholderType.alphanum
    )
    resolvable_value_property = ResolvableValueProperty('resolvable_value_property')
    value_with_none_placeholder = ResolvableValueProperty(
        'value_with_placeholder_override',
        PlaceholderType.none
    )
    config = MagicMock()


class TestResolver(object):

    def setup_method(self, test_method):
        self.mock_resolver = MockResolver(
            argument=sentinel.argument,
            stack=sentinel.stack
        )

    def test_init(self):
        assert self.mock_resolver.stack == sentinel.stack
        assert self.mock_resolver.argument == sentinel.argument


class TestResolvableContainerPropertyDescriptor:

    def setup_method(self, test_method):
        self.mock_object = MockClass()

    def test_setting_resolvable_property_with_none(self):
        self.mock_object.resolvable_container_property = None
        assert self.mock_object._resolvable_container_property is None

    def test_setting_resolvable_property_with_nested_lists(self):
        mock_resolver = MagicMock(spec=MockResolver)

        complex_data_structure = [
            "String",
            mock_resolver,
            [
                mock_resolver,
                "String",
                [
                    [
                        mock_resolver,
                        "String",
                        None
                    ],
                    mock_resolver,
                    "String"
                ]
            ]
        ]

        cloned_data_structure = [
            "String",
            mock_resolver.clone.return_value,
            [
                mock_resolver.clone.return_value,
                "String",
                [
                    [
                        mock_resolver.clone.return_value,
                        "String",
                        None
                    ],
                    mock_resolver.clone.return_value,
                    "String"
                ]
            ]
        ]

        self.mock_object.resolvable_container_property = complex_data_structure
        assert self.mock_object._resolvable_container_property == cloned_data_structure
        expected_calls = [
            call(self.mock_object),
            call().setup()
        ] * 4
        mock_resolver.clone.assert_has_calls(expected_calls)

    def test_getting_resolvable_property_with_none(self):
        self.mock_object._resolvable_container_property = None
        assert self.mock_object.resolvable_container_property is None

    def test_getting_resolvable_property_with_nested_lists(self):
        mock_resolver = MagicMock(spec=MockResolver)
        mock_resolver.resolve.return_value = "Resolved"

        complex_data_structure = [
            "String",
            mock_resolver,
            [
                mock_resolver,
                "String",
                [
                    [
                        mock_resolver,
                        "String",
                        None
                    ],
                    mock_resolver,
                    "String"
                ],
                None
            ],
            None
        ]

        resolved_complex_data_structure = [
            "String",
            "Resolved",
            [
                "Resolved",
                "String",
                [
                    [
                        "Resolved",
                        "String",
                        None
                    ],
                    "Resolved",
                    "String"
                ],
                None
            ],
            None
        ]

        self.mock_object._resolvable_container_property = complex_data_structure
        prop = self.mock_object.resolvable_container_property
        assert prop == resolved_complex_data_structure

    def test_getting_resolvable_property_with_nested_dictionaries_and_lists(
        self
    ):
        mock_resolver = MagicMock(spec=MockResolver)
        mock_resolver.resolve.return_value = "Resolved"

        complex_data_structure = {
            "String": "String",
            "None": None,
            "Resolver": mock_resolver,
            "List": [
                    [
                        mock_resolver,
                        "String",
                        None
                    ],
                    {
                        "Dictionary": {},
                        "String": "String",
                        "None": None,
                        "Resolver": mock_resolver,
                        "List": [
                            mock_resolver
                        ]
                    },
                    mock_resolver,
                    "String"
            ],
            "Dictionary": {
                "Resolver": mock_resolver,
                "Dictionary": {
                    "List": [
                        [
                            mock_resolver,
                            "String",
                            None
                        ],
                        mock_resolver,
                        "String"
                    ],
                    "String": "String",
                    "None": None,
                    "Resolver": mock_resolver
                },
                "String": "String",
                "None": None
            }
        }

        resolved_complex_data_structure = {
            "String": "String",
            "None": None,
            "Resolver": "Resolved",
            "List": [
                    [
                        "Resolved",
                        "String",
                        None
                    ],
                    {
                        "Dictionary": {},
                        "String": "String",
                        "None": None,
                        "Resolver": "Resolved",
                        "List": [
                            "Resolved"
                        ]
                    },
                    "Resolved",
                    "String"
            ],
            "Dictionary": {
                "Resolver": "Resolved",
                "Dictionary": {
                    "List": [
                        [
                            "Resolved",
                            "String",
                            None
                        ],
                        "Resolved",
                        "String"
                    ],
                    "String": "String",
                    "None": None,
                    "Resolver": "Resolved"
                },
                "String": "String",
                "None": None
            }
        }

        self.mock_object._resolvable_container_property = complex_data_structure
        prop = self.mock_object.resolvable_container_property
        assert prop == resolved_complex_data_structure

    def test_getting_resolvable_property_with_nested_dictionaries(self):
        mock_resolver = MagicMock(spec=MockResolver)
        mock_resolver.resolve.return_value = "Resolved"

        complex_data_structure = {
            "String": "String",
            "None": None,
            "Resolver": mock_resolver,
            "Dictionary": {
                "Resolver": mock_resolver,
                "Dictionary": {
                    "Dictionary": {},
                    "String": "String",
                    "None": None,
                    "Resolver": mock_resolver
                },
                "String": "String",
                "None": None
            }
        }

        resolved_complex_data_structure = {
            "String": "String",
            "None": None,
            "Resolver": "Resolved",
            "Dictionary": {
                "Resolver": "Resolved",
                "Dictionary": {
                    "Dictionary": {},
                    "String": "String",
                    "None": None,
                    "Resolver": "Resolved"
                },
                "String": "String",
                "None": None
            }
        }

        self.mock_object._resolvable_container_property = complex_data_structure
        prop = self.mock_object.resolvable_container_property
        assert prop == resolved_complex_data_structure

    def test_get__resolver_references_same_property_for_other_value__resolves_it(self):
        class MyResolver(Resolver):
            def resolve(self):
                return self.stack.resolvable_container_property['other_value']

        resolver = MyResolver()
        self.mock_object.resolvable_container_property = {
            'other_value': 'abc',
            'resolver': resolver
        }

        assert self.mock_object.resolvable_container_property['resolver'] == 'abc'

    def test_get__resolver_references_itself__raises_recursive_resolve(self):
        class RecursiveResolver(Resolver):
            def resolve(self):
                return self.stack.resolvable_container_property['resolver']

        resolver = RecursiveResolver()
        self.mock_object.resolvable_container_property = {
            'resolver': resolver
        }
        with pytest.raises(RecursiveResolve):
            self.mock_object.resolvable_container_property

    def test_get__resolvable_container_property_references_same_property_of_other_stack__resolves(self):
        stack1 = MockClass()
        stack1.resolvable_container_property = {
            'testing': 'stack1'
        }

        class OtherStackResolver(Resolver):
            def resolve(self):
                return stack1.resolvable_container_property['testing']

        stack2 = MockClass()
        stack2.resolvable_container_property = {
            'resolver': OtherStackResolver()
        }

        assert stack2.resolvable_container_property == {
            'resolver': 'stack1'
        }

    def test_get__resolver_resolves_to_none__value_is_dict__deletes_those_items_from_dict(self):
        class MyResolver(Resolver):
            def resolve(self):
                return None

        resolver = MyResolver()
        self.mock_object.resolvable_container_property = {
            'a': 4,
            'b': resolver,
            'c': 3,
            'd': resolver,
            'e': resolver,
            'f': 5,
        }
        expected = {'a': 4, 'c': 3, 'f': 5}
        assert self.mock_object.resolvable_container_property == expected

    def test_get__resolver_resolves_to_none__value_is_dict__deletes_those_items_from_complex_structure(self):
        class MyResolver(Resolver):
            def resolve(self):
                return None

        resolver = MyResolver()
        self.mock_object.resolvable_container_property = {
            'a': 4,
            'b': [
                resolver,
            ],
            'c': [{
                'v': resolver
            }],
            'd': 3
        }
        expected = {'a': 4, 'b': [], 'c': [{}], 'd': 3}
        assert self.mock_object.resolvable_container_property == expected

    def test_get__resolver_resolves_to_none__value_is_list__deletes_that_item_from_list(self):
        class MyResolver(Resolver):
            def resolve(self):
                return None

        resolver = MyResolver()
        self.mock_object.resolvable_container_property = [
            1,
            resolver,
            3
        ]
        expected = [1, 3]
        assert self.mock_object.resolvable_container_property == expected

    def test_get__resolver_resolves_to_none__value_is_dict__deletes_that_key_from_dict(self):
        class MyResolver(Resolver):
            def resolve(self):
                return None

        resolver = MyResolver()
        self.mock_object.resolvable_container_property = {
            'some key': 'some value',
            'resolver': resolver
        }
        expected = {'some key': 'some value'}
        assert self.mock_object.resolvable_container_property == expected

    def test_get__resolvers_resolves_to_none__value_is_list__deletes_those_items_from_list(self):
        class MyResolver(Resolver):
            def resolve(self):
                return None

        resolver = MyResolver()

        self.mock_object.resolvable_container_property = [
            1,
            resolver,
            3,
            resolver,
            resolver,
            6
        ]
        expected = [1, 3, 6]
        assert self.mock_object.resolvable_container_property == expected

    def test_get__resolvers_resolves_to_none__value_is_list__deletes_all_items_from_list(self):
        class MyResolver(Resolver):
            def resolve(self):
                return None

        resolver = MyResolver()

        self.mock_object.resolvable_container_property = [
            resolver,
            resolver,
            resolver
        ]
        expected = []
        assert self.mock_object.resolvable_container_property == expected

    def test_get__value_in_list_is_none__returns_list_with_none(self):
        self.mock_object.resolvable_container_property = [
            1,
            None,
            3
        ]
        expected = [1, None, 3]
        assert self.mock_object.resolvable_container_property == expected

    def test_get__value_in_dict_is_none__returns_dict_with_none(self):
        self.mock_object.resolvable_container_property = {
            'some key': 'some value',
            'none key': None
        }
        expected = {'some key': 'some value', 'none key': None}
        assert self.mock_object.resolvable_container_property == expected

    def test_get__resolver_raises_error__placeholders_allowed__returns_placeholder(self):
        class ErroringResolver(Resolver):
            def resolve(self):
                raise ValueError()

        resolver = ErroringResolver()
        self.mock_object.resolvable_container_property = {
            'resolver': resolver
        }
        with use_resolver_placeholders_on_error():
            result = self.mock_object.resolvable_container_property

        assert result == {'resolver': create_placeholder_value(resolver, PlaceholderType.explicit)}

    def test_get__resolver_raises_error__placeholders_not_allowed__raises_error(self):
        class ErroringResolver(Resolver):
            def resolve(self):
                raise ValueError()

        resolver = ErroringResolver()
        self.mock_object.resolvable_container_property = {
            'resolver': resolver
        }
        with pytest.raises(ValueError):
            self.mock_object.resolvable_container_property

    def test_get__resolver_raises_recursive_resolve__placeholders_allowed__raises_error(self):
        class RecursiveResolver(Resolver):
            def resolve(self):
                raise RecursiveResolve()

        resolver = RecursiveResolver()
        self.mock_object.resolvable_container_property = {
            'resolver': resolver
        }
        with use_resolver_placeholders_on_error(), pytest.raises(RecursiveResolve):
            self.mock_object.resolvable_container_property

    def test_get__resolver_raises_error__placeholders_allowed__alternate_placeholder_type__uses_alternate(self):
        class ErroringResolver(Resolver):
            def resolve(self):
                raise ValueError()

        resolver = ErroringResolver()
        self.mock_object.container_with_alphanum_placeholder = {
            'resolver': resolver
        }
        with use_resolver_placeholders_on_error():
            result = self.mock_object.container_with_alphanum_placeholder

        assert result == {'resolver': create_placeholder_value(resolver, PlaceholderType.alphanum)}


class TestResolvableValueProperty:
    def setup_method(self, test_method):
        self.mock_object = MockClass()

    @pytest.mark.parametrize(
        'value',
        ['string', True, 123, 1.23, None]
    )
    def test_set__non_resolver__sets_private_variable_as_value(self, value):
        self.mock_object.resolvable_value_property = value
        assert self.mock_object._resolvable_value_property == value

    def test_set__resolver__sets_private_variable_with_clone_of_resolver_with_instance(self):
        resolver = Mock(spec=MockResolver)
        self.mock_object.resolvable_value_property = resolver
        assert self.mock_object._resolvable_value_property == resolver.clone.return_value

    def test_set__resolver__sets_up_cloned_resolver(self):
        resolver = Mock(spec=MockResolver)
        self.mock_object.resolvable_value_property = resolver
        resolver.clone.return_value.setup.assert_any_call()

    @pytest.mark.parametrize(
        'value',
        ['string', True, 123, 1.23, None]
    )
    def test_get__non_resolver__returns_value(self, value):
        self.mock_object._resolvable_value_property = value
        assert self.mock_object.resolvable_value_property == value

    def test_get__resolver__returns_resolved_value(self):
        resolver = Mock(spec=MockResolver)
        self.mock_object._resolvable_value_property = resolver
        assert self.mock_object.resolvable_value_property == resolver.resolve.return_value

    def test_get__resolver__updates_set_value_with_resolved_value(self):
        resolver = Mock(spec=MockResolver)
        self.mock_object._resolvable_value_property = resolver
        self.mock_object.resolvable_value_property
        assert self.mock_object._resolvable_value_property == resolver.resolve.return_value

    def test_get__resolver__resolver_attempts_to_access_resolver__raises_recursive_resolve(self):
        class RecursiveResolver(Resolver):
            def resolve(self):
                # This should blow up!
                self.stack.resolvable_value_property

        resolver = RecursiveResolver()
        self.mock_object.resolvable_value_property = resolver

        with pytest.raises(RecursiveResolve):
            self.mock_object.resolvable_value_property

    def test_get__resolvable_value_property_references_same_property_of_other_stack__resolves(self):
        stack1 = MockClass()
        stack1.resolvable_value_property = 'stack1'

        class OtherStackResolver(Resolver):
            def resolve(self):
                return stack1.resolvable_value_property

        stack2 = MockClass()
        stack2.resolvable_value_property = OtherStackResolver()

        assert stack2.resolvable_value_property == 'stack1'

    def test_get__resolver_raises_error__placeholders_allowed__returns_placeholder(self):
        class ErroringResolver(Resolver):
            def resolve(self):
                raise ValueError()

        resolver = ErroringResolver()
        self.mock_object.resolvable_value_property = resolver
        with use_resolver_placeholders_on_error():
            result = self.mock_object.resolvable_value_property

        assert result == create_placeholder_value(resolver, PlaceholderType.explicit)

    def test_get__resolver_raises_error__placeholders_not_allowed__raises_error(self):
        class ErroringResolver(Resolver):
            def resolve(self):
                raise ValueError()

        resolver = ErroringResolver()
        self.mock_object.resolvable_value_property = resolver
        with pytest.raises(ValueError):
            self.mock_object.resolvable_value_property

    def test_get__resolver_raises_recursive_resolve__placeholders_allowed__raises_error(self):
        class RecursiveResolver(Resolver):
            def resolve(self):
                raise RecursiveResolve()

        resolver = RecursiveResolver()
        self.mock_object.resolvable_value_property = resolver
        with use_resolver_placeholders_on_error(), pytest.raises(RecursiveResolve):
            self.mock_object.resolvable_value_property

    def test_get__resolver_raises_error__placeholders_allowed__alternate_placeholder_type__uses_alternate_type(self):
        class ErroringResolver(Resolver):
            def resolve(self):
                raise ValueError()

        resolver = ErroringResolver()
        self.mock_object.value_with_none_placeholder = resolver
        with use_resolver_placeholders_on_error():
            result = self.mock_object.value_with_none_placeholder

        assert result == create_placeholder_value(resolver, PlaceholderType.none)



import io
import sys
from logging import getLogger
from typing import Dict, TextIO, Type, Iterable

import click
from click import Context

from sceptre.cli.helpers import catch_exceptions
from sceptre.context import SceptreContext
from sceptre.diffing.diff_writer import DeepDiffWriter, DiffLibWriter, DiffWriter
from sceptre.diffing.stack_differ import DeepDiffStackDiffer, DifflibStackDiffer, StackDiff
from sceptre.helpers import null_context
from sceptre.plan.plan import SceptrePlan
from sceptre.resolvers.placeholders import use_resolver_placeholders_on_error
from sceptre.stack import Stack

logger = getLogger(__name__)


@click.command(name="diff", short_help="Compares deployed infrastructure with current configurations")
@click.option(
    '-t',
    '--type',
    'differ',
    type=click.Choice(['deepdiff', 'difflib']),
    default='deepdiff',
    help='The type of differ to use. Use "deepdiff" for recursive key/value comparison. "difflib" '
         'produces a more traditional "diff" result. Defaults to deepdiff.'
)
@click.option(
    '-s',
    '--show-no-echo',
    is_flag=True,
    help='If set, will display the unmasked values of NoEcho parameters generated LOCALLY (NoEcho '
         'parameters for deployed stacks will always be masked when retrieved from CloudFormation.). '
         'If not set (the default), parameters identified as NoEcho on the local template will be '
         'masked when presented in the diff.'
)
@click.option(
    '-n',
    '--no-placeholders',
    is_flag=True,
    help="If set, no placeholder values will be supplied for resolvers that cannot be resolved."
)
@click.option(
    '-a',
    '--all',
    'all_',
    is_flag=True,
    help=(
         "If set, will perform diffing on ALL stacks, including ignored and obsolete ones; Otherwise, "
         "it will diff only stacks that would be created or updated when running the launch command."
    )
)
@click.argument('path')
@click.pass_context
@catch_exceptions
def diff_command(
    ctx: Context,
    differ: str,
    show_no_echo: bool,
    no_placeholders: bool,
    all_: bool,
    path: str
):
    """Indicates the difference between the currently DEPLOYED stacks in the command path and
    the stacks configured in Sceptre right now. This command will compare both the templates as well
    as the subset of stack configurations that can be compared. By default, only stacks that would
    be launched via the launch command will be diffed, but you can diff ALL stacks relevant to the
    passed command path if you pass the --all flag.

    Some settings (such as sceptre_user_data) are not available in a CloudFormation stack
    description, so the diff will not be indicated. Currently compared stack configurations are:

    \b
      * parameters
      * notifications
      * role_arn
      * stack_tags

    Important: There are resolvers (notably !stack_output) that rely on other stacks
    to be already deployed when they are resolved. When producing a diff on Stack Configs that have
    such resolvers that point to non-deployed stacks, this presents a challenge, since this means
    those resolvers cannot be resolved. This particularly applies to stack parameters and when a
    stack's template uses sceptre_user_data with resolvers in it. In order to continue to be useful
    when producing a diff in these conditions, this command will do the following:

    1. If the resolver CAN be resolved, it will be resolved and the resolved value will be in the
    diff results.
    2. If the resolver CANNOT be resolved, it will be replaced with a string that represents the
    resolver and its arguments. For example: !stack_output my_stack.yaml::MyOutput will resolve in
    the parameters to "{ !StackOutput(my_stack.yaml::MyOutput) }".

    Particularly in cases where the replaced value doesn't work in the template as the template logic
    requires and causes an error, there is nothing further Sceptre can do and diffing will fail.
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
        output_format=ctx.obj.get('output_format'),
        no_colour=ctx.obj.get('no_colour')
    )
    output_format = context.output_format
    plan = SceptrePlan(context)
    if not all_:
        filter_plan_for_launchable(plan)

    if differ == "deepdiff":
        stack_differ = DeepDiffStackDiffer(show_no_echo)
        writer_class = DeepDiffWriter
    elif differ == 'difflib':
        stack_differ = DifflibStackDiffer(show_no_echo)
        writer_class = DiffLibWriter
    else:
        raise ValueError(f"Unexpected differ type: {differ}")

    execution_context = null_context() if no_placeholders else use_resolver_placeholders_on_error()
    with execution_context:
        diffs: Dict[Stack, StackDiff] = plan.diff(stack_differ)

    num_stacks_with_diff = output_diffs(diffs.values(), writer_class, sys.stdout, output_format)

    if num_stacks_with_diff:
        logger.warning(
            f"{num_stacks_with_diff} stacks with differences detected."
        )


def output_diffs(
    diffs: Iterable[StackDiff],
    writer_class: Type[DiffWriter],
    output_buffer: TextIO,
    output_format: str,
) -> int:
    """Outputs the diff results to the output_buffer.

    :param diffs: The differences computed
    :param writer_class: The DiffWriter class to be instantiated for each StackDiff
    :param output_buffer: The buffer to write the diff results to
    :param output_format: The format to output the results in
    :return: The number of stacks that had a difference
    """

    line_buffer = io.StringIO()

    num_stacks_with_diff = 0

    for stack_diff in diffs:
        writer = writer_class(stack_diff, line_buffer, output_format)
        writer.write()
        if writer.has_difference:
            num_stacks_with_diff += 1

    output_buffer_with_normalized_bar_lengths(line_buffer, output_buffer)
    return num_stacks_with_diff


def output_buffer_with_normalized_bar_lengths(buffer: io.StringIO, output_stream: TextIO):
    """Takes the output from a buffer and ensures that the star and line bars are the same length
    across the entire buffer and that their length is the full width of longest line.

    :param buffer: The input stream to normalize bar lengths for
    :param output_stream: The stream to output the normalized buffer into
    """
    buffer.seek(0)
    max_length = len(max(buffer, key=len))
    buffer.seek(0)
    full_length_star_bar = '*' * max_length
    full_length_line_bar = '-' * max_length
    for line in buffer:
        if DiffWriter.STAR_BAR in line:
            line = line.replace(DiffWriter.STAR_BAR, full_length_star_bar)
        if DiffWriter.LINE_BAR in line:
            line = line.replace(DiffWriter.LINE_BAR, full_length_line_bar)
        output_stream.write(line)


def filter_plan_for_launchable(plan: SceptrePlan):
    plan.resolve(plan.diff.__name__)
    plan.filter(lambda stack: not stack.ignore and not stack.obsolete)



# -*- coding: utf-8 -*-
import abc
import logging
from contextlib import contextmanager
from threading import RLock
from typing import Any, TYPE_CHECKING, Type, Union, TypeVar

from sceptre.helpers import _call_func_on_values
from sceptre.resolvers.placeholders import (
    create_placeholder_value,
    are_placeholders_enabled, PlaceholderType
)

if TYPE_CHECKING:
    from sceptre import stack

T_Container = TypeVar('T_Container', bound=Union[dict, list])


class RecursiveResolve(Exception):
    pass


class Resolver(abc.ABC):
    """
    Resolver is an abstract base class that should be inherited by all
    Resolvers.

    :param argument: The argument of the resolver.
    :param stack: The associated stack of the resolver.
    """

    def __init__(self, argument: Any = None, stack: 'stack.Stack' = None):
        self.logger = logging.getLogger(__name__)
        self.argument = argument
        self.stack = stack

    def setup(self):
        """
        This method is called at during stack initialisation.
        Implementation of this method in subclasses can be used to do any
        initial setup of the object.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def resolve(self):
        """
        An abstract method which must be overwritten by all inheriting classes.
        This method is called to retrieve the final desired value.
        Implementation of this method in subclasses must return a suitable
        object or primitive type.
        """
        pass  # pragma: no cover

    def clone(self, stack: 'stack.Stack') -> 'Resolver':
        """
        Produces a "fresh" copy of the Resolver, with the specified stack.

        :param stack: The stack to set on the cloned resolver
        """
        return type(self)(self.argument, stack)


class ResolvableProperty(abc.ABC):
    """
    This is an abstract base class for a descriptor used to store an attribute that have values
    associated with Resolver objects.

    :param name: Attribute suffix used to store the property in the instance.
    :param placeholder_type: The type of placeholder that should be returned, when placeholders are
        allowed, when a resolver can't be resolved.
    """

    def __init__(self, name: str, placeholder_type=PlaceholderType.explicit):
        self.name = "_" + name
        self.logger = logging.getLogger(__name__)
        self.placeholder_type = placeholder_type

        self._lock = RLock()

    def __get__(self, stack: 'stack.Stack', stack_class: Type['stack.Stack']) -> Any:
        """
        Attribute getter which resolves the resolver(s).

        :param stack: The Stack instance the property is being retrieved for
        :param stack_class: The class of the stack that the property is being retrieved for.
        :return: The attribute stored with the suffix ``name`` in the instance.
        :rtype: The obtained value, as resolved by the property
        """
        with self._lock, self._no_recursive_get(stack):
            if hasattr(stack, self.name):
                return self.get_resolved_value(stack, stack_class)

    def __set__(self, stack: 'stack.Stack', value: Any):
        """
        Attribute setter which adds a stack reference to any resolvers in the
        data structure `value` and calls the setup method.

        :param stack: The Stack instance the value is being set onto
        :param value: The value being set on the property
        """
        with self._lock:
            self.assign_value_to_stack(stack, value)

    @contextmanager
    def _no_recursive_get(self, stack: 'stack.Stack'):
        # We don't care about recursive gets on the same property but different Stack instances,
        # only recursive gets on the same stack. Some Resolvers access the same property on OTHER
        # stacks and that actually shouldn't be a problem. Remember, these descriptor instances are
        # set on the CLASS and so instance variables on them are shared across all classes that
        # access them. Thus, we set this "get_in_progress" attribute on the stack instance rather
        # than the descriptor instance.
        get_status_name = f'_{self.name}_get_in_progress'
        if getattr(stack, get_status_name, False):
            raise RecursiveResolve(f"Resolving Stack.{self.name[1:]} required resolving itself")
        setattr(stack, get_status_name, True)
        try:
            yield
        finally:
            setattr(stack, get_status_name, False)

    def get_setup_resolver_for_stack(self, stack: 'stack.Stack', resolver: Resolver) -> Resolver:
        """Obtains a clone of the resolver with the stack set on it and the setup method having
        been called on it.

        :param stack: The stack to set on the Resolver
        :param resolver: The Resolver to clone and set up
        :return: The cloned resolver.
        """
        # We clone the resolver when we assign the value so that every stack gets its own resolver
        # rather than potentially having one resolver instance shared in memory across multiple
        # stacks.
        clone = resolver.clone(stack)
        clone.setup()
        return clone

    @abc.abstractmethod
    def get_resolved_value(self, stack: 'stack.Stack', stack_class: Type['stack.Stack']) -> Any:
        """Implement this method to return the value of the resolvable_property."""
        pass

    @abc.abstractmethod
    def assign_value_to_stack(self, stack: 'stack.Stack', value: Any):
        """Implement this method to assign the value to the resolvable property."""
        pass

    def resolve_resolver_value(self, resolver: 'Resolver') -> Any:
        """Returns the resolved parameter value.

        If the resolver happens to raise an error and placeholders are currently allowed for resolvers,
        a placeholder will be returned instead of reraising the error.

        :param resolver: The resolver to resolve.
        :return: The resolved value (or placeholder, in certain circumstances)
        """
        try:
            return resolver.resolve()
        except RecursiveResolve:
            # Recursive resolve issues shouldn't be masked by a placeholder.
            raise
        except Exception:
            if are_placeholders_enabled():
                placeholder_value = create_placeholder_value(resolver, self.placeholder_type)

                self.logger.debug(
                    "Error encountered while resolving the resolver. This is allowed for the current "
                    f"operation. Resolving it to a placeholder value instead: {placeholder_value}"
                )
                return placeholder_value
            raise

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}({self.name[1:]})>'


class ResolvableContainerProperty(ResolvableProperty):
    """
    This is a descriptor class used to store an attribute that may CONTAIN
    Resolver objects. Resolvers will be resolved upon access of this property.
    When resolvers are resolved, they will be replaced in the container with their
    resolved value, in order to avoid redundant resolutions.

    Supports nested dictionary and lists.

    :param name: Attribute suffix used to store the property in the instance.
    :type name: str
    """

    def __get__(self, stack: 'stack.Stack', stack_class: Type['stack.Stack']) -> T_Container:
        container = super().__get__(stack, stack_class)

        with self._lock:
            # Resolve any deferred resolvers, now that the recursive get lock has been released.
            self._resolve_deferred_resolvers(stack, container)

        return container

    def get_resolved_value(self, stack: 'stack.Stack', stack_class: Type['stack.Stack']) -> T_Container:
        """Obtains the resolved value for this property. Any resolvers that resolve to None will have
        their key/index removed from their dict/list where they are. Other resolvers will have their
        key/index's value replace with the resolved value to avoid redundant resolutions.

        :param stack: The Stack instance to obtain the value for
        :param stack_class: The class of the Stack instance.
        :return: The fully resolved container.
        """
        keys_to_delete = []

        def resolve(attr: Union[dict, list], key: Union[int, str], value: Resolver):
            # Update the container key's value with the resolved value, if possible...
            try:
                result = self.resolve_resolver_value(value)
                if result is None:
                    self.logger.debug(f"Removing item {key} because resolver returned None.")
                    # We gather up resolvers (and their immediate containers) that resolve to None,
                    # since that really means the resolver resolves to nothing. This is not common,
                    # but should be supported. We gather these rather than immediately remove them
                    # because this function is called in the context of looping over that attr, so
                    # we cannot alter its size until after the loop is complete.
                    keys_to_delete.append((attr, key))
                else:
                    attr[key] = result
            except RecursiveResolve:
                # It's possible that resolving the resolver might attempt to access another
                # resolvable property's value in this same container. In this case, we'll delay
                # resolution and instead return a ResolveLater so the value can be resolved outside
                # this recursion.
                attr[key] = self.ResolveLater(
                    stack,
                    self.name,
                    key,
                    lambda: value.resolve(),
                )

        container = getattr(stack, self.name)
        _call_func_on_values(
            resolve, container, Resolver
        )
        # Remove keys and indexes from their containers that had resolvers resolve to None.
        list_items_to_delete = []
        for attr, key in keys_to_delete:
            if isinstance(attr, list):
                # If it's a list, we want to gather up the items to remove from the list.
                # We don't want to modify the list length yet.
                # Since removals will change all the other list indexes,
                # we don't wan't to modify lists yet.
                list_items_to_delete.append((attr, attr[key]))
            else:
                del attr[key]

        for containing_list, item in list_items_to_delete:
            containing_list.remove(item)

        return container

    def assign_value_to_stack(self, stack: 'stack.Stack', value: Union[dict, list]):
        """Assigns a COPY of the specified value to the stack instance. This method copies the value
        rather than directly assigns it to avoid bugs related to shared objects in memory.

        :param stack: The stack to assign the value to
        :param value: The value to assign
        """
        cloned = self._clone_container_with_resolvers(value, stack)
        setattr(stack, self.name, cloned)

    def _clone_container_with_resolvers(
        self,
        container: T_Container,
        stack: 'stack.Stack'
    ) -> T_Container:
        """Recurses into the container, cloning and setting up resolvers and creating a copy of all
        nested containers.

        :param container: The container being recursed into and cloned
        :param stack: The stack the container is being copied for
        :return: The fully copied container with resolvers fully set up.
        """
        def recurse(obj):
            if isinstance(obj, Resolver):
                return self.get_setup_resolver_for_stack(stack, obj)
            if isinstance(obj, list):
                return [
                    recurse(item)
                    for item in obj
                ]
            elif isinstance(obj, dict):
                return {
                    key: recurse(val)
                    for key, val in obj.items()
                }
            return obj

        return recurse(container)

    def _resolve_deferred_resolvers(self, stack: 'stack.Stack', container: T_Container):
        def raise_if_not_resolved(attr, key, value):
            # If this function has been hit, it means that after attempting to resolve all the
            # ResolveLaters, there STILL are ResolveLaters left in the container. Rather than
            # continuing to try to resolve (possibly infinitely), we'll raise a RecursiveGet to
            # break that infinite loop. This situation would happen if a resolver accesses a resolver
            # in the same container, which then accesses another resolver (possibly the same one) in
            # the same container.
            raise RecursiveResolve(f"Resolving Stack.{self.name[1:]} required resolving itself")

        has_been_resolved_attr_name = f'{self.name}_is_resolved'
        if not getattr(stack, has_been_resolved_attr_name, False):
            # We set it first rather than after to avoid entering this block again on this property
            # for this stack.
            setattr(stack, has_been_resolved_attr_name, True)
            _call_func_on_values(
                lambda attr, key, value: value(),
                container,
                self.ResolveLater
            )
            # Search the container to see if there are any ResolveLaters left;
            # Raise a RecursiveResolve if there are.
            _call_func_on_values(
                raise_if_not_resolved,
                container,
                self.ResolveLater
            )

    class ResolveLater:
        """Represents a value that could not yet be resolved but can be resolved in the future."""

        def __init__(self, instance, name, key, resolution_function):
            self._instance = instance
            self._name = name
            self._key = key
            self._resolution_function = resolution_function

        def __call__(self):
            """Resolve the value."""
            attr = getattr(self._instance, self._name)
            result = self._resolution_function()
            if result is None:
                self.logger.debug(f"Removing item {self._key} because resolver returned None.")
                del attr[self._key]
            else:
                attr[self._key] = result


class ResolvableValueProperty(ResolvableProperty):
    """
    This is a descriptor class used to store an attribute that may BE a single
    Resolver object. If it is a resolver, it will be resolved upon access of this property.
    When resolved, the resolved value will replace the resolver on the stack in order to avoid
    redundant resolutions.

    :param name: Attribute suffix used to store the property in the instance.
    :type name: str
    """

    def get_resolved_value(self, stack: 'stack.Stack', stack_class: Type['stack.Stack']) -> Any:
        """Gets the fully-resolved value from the property. Resolvers will be replaced on the stack
        instance with their resolved value to avoid redundant resolutions.

        :param stack: The Stack instance to obtain the value from
        :param stack_class: The class of the Stack instance
        :return: The fully resolved value
        """
        raw_value = getattr(stack, self.name)
        if isinstance(raw_value, Resolver):
            value = self.resolve_resolver_value(raw_value)
            # Overwrite the stored resolver value with the resolved value to avoid resolving the
            # same value multiple times.
            setattr(stack, self.name, value)
        else:
            value = raw_value

        return value

    def assign_value_to_stack(self, stack: 'stack.Stack', value: Any):
        """Assigns the value to the Stack instance passed, setting up and cloning the value if it
        is a Resolver.

        :param stack: The Stack instance to set the value on
        :param value: The value to set
        """
        if isinstance(value, Resolver):
            value = self.get_setup_resolver_for_stack(stack, value)
        setattr(stack, self.name, value)



import logging
import click

from sceptre.context import SceptreContext
from sceptre.cli.helpers import (
    catch_exceptions,
    write
)
from sceptre.plan.plan import SceptrePlan

logger = logging.getLogger(__name__)


@click.group(name="list")
def list_group():
    """
    Commands for listing attributes of stacks.
    """
    pass


@list_group.command(name="resources")
@click.argument("path")
@click.pass_context
@catch_exceptions
def list_resources(ctx, path):
    """
    List resources for stack or stack_group.
    \f

    :param path: Path to execute the command on.
    :type path: str
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies")
    )
    plan = SceptrePlan(context)

    responses = [
        response for response
        in plan.describe_resources().values() if response
    ]

    write(responses, context.output_format)


@list_group.command(name="outputs")
@click.argument("path")
@click.option(
    "-e", "--export", type=click.Choice(["envvar"]),
    help="Specify the export formatting."
)
@click.pass_context
@catch_exceptions
def list_outputs(ctx, path, export):
    """
    List outputs for stack.
    \f

    :param path: Path to execute the command on.
    :type path: str
    :param export: Specify the export formatting.
    :type export: str
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path", None),
        user_variables=ctx.obj.get("user_variables", {}),
        options=ctx.obj.get("options", {}),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies")
    )

    plan = SceptrePlan(context)
    responses = [
        response for response
        in plan.describe_outputs().values() if response
    ]

    if export == "envvar":
        for response in responses:
            for stack in response.values():
                for output in stack:
                    write("export SCEPTRE_{0}='{1}'".format(
                        output.get("OutputKey"),
                        output.get("OutputValue")
                    ), 'text')
    else:
        write(responses, context.output_format)


@list_group.command(name="change-sets")
@click.option(
    "-U", "--url", is_flag=True, help="Instead write a URL."
)
@click.argument("path")
@click.pass_context
@catch_exceptions
def list_change_sets(ctx, path, url):
    """
    List change sets for stack.
    \f

    :param path: Path to execute the command on.
    :type path: str
    :param url: Write out a console URL instead.
    :type url: bool
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        output_format=ctx.obj.get("output_format"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies")
    )

    plan = SceptrePlan(context)

    responses = [
        response for response
        in plan.list_change_sets(url).values() if response
    ]

    for response in responses:
        write(response, context.output_format)


@list_group.command(name="stacks")
@click.argument("path")
@click.pass_context
@catch_exceptions
def list_stacks(ctx, path):
    """
    List sceptre stack config attributes,
    \f

    :param path: Path to execute the command on or path to stack group
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        output_format=ctx.obj.get("output_format"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies")
    )

    plan = SceptrePlan(context)

    output = {f"{stack.name}.yaml": stack.external_name for stack in plan.graph}
    output_format = "json" if context.output_format == "json" else "yaml"
    write(output, output_format)



from unittest.mock import Mock

import pytest

from sceptre.resolvers.stack_attr import StackAttr
from sceptre.stack import Stack


class TestResolver(object):

    def setup_method(self, test_method):
        self.stack_group_config = {}
        self.stack = Mock(spec=Stack, stack_group_config=self.stack_group_config)

        self.resolver = StackAttr(stack=self.stack)

    def test__resolve__returns_attribute_off_stack(self):
        self.resolver.argument = 'testing_this'
        self.stack.testing_this = 'hurray!'
        result = self.resolver.resolve()

        assert result == 'hurray!'

    def test_resolve__nested_attribute__accesses_nested_value(self):
        self.stack.testing_this = {
            'top': [
                {'thing': 'first'},
                {'thing': 'second'}
            ]
        }

        self.resolver.argument = 'testing_this.top.1.thing'
        result = self.resolver.resolve()

        assert result == 'second'

    def test_resolve__attribute_not_defined__accesses_it_off_stack_group_config(self):
        self.stack.stack_group_config['testing_this'] = {
            'top': [
                {'thing': 'first'},
                {'thing': 'second'}
            ]
        }

        self.resolver.argument = 'testing_this.top.1.thing'
        result = self.resolver.resolve()

        assert result == 'second'

    @pytest.mark.parametrize('config,attr_name', [
        ('template', 'template_handler_config'),
        ('protect', 'protected'),
        ('stack_name', 'external_name'),
        ('stack_tags', 'tags')
    ])
    def test_resolve__accessing_attribute_renamed_on_stack__resolves_correct_value(self, config, attr_name):
        setattr(self.stack, attr_name, 'value')
        self.resolver.argument = config

        result = self.resolver.resolve()

        assert result == 'value'

    def test_resolve__attribute_not_defined__raises_attribute_error(self):
        self.resolver.argument = 'nonexistant'

        with pytest.raises(AttributeError):
            self.resolver.resolve()



# -*- coding: utf-8 -*-

from unittest.mock import MagicMock, sentinel

import pytest

from sceptre.exceptions import InvalidConfigFileError
from sceptre.resolvers import Resolver
from sceptre.stack import Stack
from sceptre.template import Template


def stack_factory(**kwargs):
    call_kwargs = {
        'name': 'dev/app/stack',
        'project_code': sentinel.project_code,
        'template_bucket_name': sentinel.template_bucket_name,
        'template_key_prefix': sentinel.template_key_prefix,
        'required_version': sentinel.required_version,
        'template_path': sentinel.template_path,
        'region': sentinel.region,
        'profile': sentinel.profile,
        'parameters': {"key1": "val1"},
        'sceptre_user_data': sentinel.sceptre_user_data,
        'hooks': {},
        's3_details': None,
        'dependencies': sentinel.dependencies,
        'role_arn': sentinel.role_arn,
        'protected': False,
        'tags': {"tag1": "val1"},
        'external_name': sentinel.external_name,
        'notifications': [sentinel.notification],
        'on_failure': sentinel.on_failure,
        'stack_timeout': sentinel.stack_timeout,
        'stack_group_config': {}
    }
    call_kwargs.update(kwargs)
    return Stack(**call_kwargs)


class TestStack(object):

    def setup_method(self, test_method):
        self.stack = Stack(
            name='dev/app/stack', project_code=sentinel.project_code,
            template_bucket_name=sentinel.template_bucket_name,
            template_key_prefix=sentinel.template_key_prefix,
            required_version=sentinel.required_version,
            template_path=sentinel.template_path,
            region=sentinel.region,
            profile=sentinel.profile, parameters={"key1": "val1"},
            sceptre_user_data=sentinel.sceptre_user_data, hooks={},
            s3_details=None, dependencies=sentinel.dependencies,
            role_arn=sentinel.role_arn, protected=False,
            tags={"tag1": "val1"}, external_name=sentinel.external_name,
            notifications=[sentinel.notification],
            on_failure=sentinel.on_failure, iam_role=sentinel.iam_role,
            iam_role_session_duration=sentinel.iam_role_session_duration,
            stack_timeout=sentinel.stack_timeout,
            stack_group_config={}
        )
        self.stack._template = MagicMock(spec=Template)

    def test_initialize_stack_with_template_path(self):
        stack = Stack(
            name='dev/stack/app', project_code=sentinel.project_code,
            template_path=sentinel.template_path,
            template_bucket_name=sentinel.template_bucket_name,
            template_key_prefix=sentinel.template_key_prefix,
            required_version=sentinel.required_version,
            region=sentinel.region, external_name=sentinel.external_name
        )
        assert stack.name == 'dev/stack/app'
        assert stack.project_code == sentinel.project_code
        assert stack.template_bucket_name == sentinel.template_bucket_name
        assert stack.template_key_prefix == sentinel.template_key_prefix
        assert stack.required_version == sentinel.required_version
        assert stack.external_name == sentinel.external_name
        assert stack.hooks == {}
        assert stack.parameters == {}
        assert stack.sceptre_user_data == {}
        assert stack.template_path == sentinel.template_path
        assert stack.template_handler_config is None
        assert stack.s3_details is None
        assert stack._template is None
        assert stack.protected is False
        assert stack.iam_role is None
        assert stack.role_arn is None
        assert stack.dependencies == []
        assert stack.tags == {}
        assert stack.notifications == []
        assert stack.on_failure is None
        assert stack.stack_group_config == {}

    def test_initialize_stack_with_template_handler(self):
        stack = Stack(
            name='dev/stack/app', project_code=sentinel.project_code,
            template_handler_config=sentinel.template_handler_config,
            template_bucket_name=sentinel.template_bucket_name,
            template_key_prefix=sentinel.template_key_prefix,
            required_version=sentinel.required_version,
            region=sentinel.region, external_name=sentinel.external_name
        )
        assert stack.name == 'dev/stack/app'
        assert stack.project_code == sentinel.project_code
        assert stack.template_bucket_name == sentinel.template_bucket_name
        assert stack.template_key_prefix == sentinel.template_key_prefix
        assert stack.required_version == sentinel.required_version
        assert stack.external_name == sentinel.external_name
        assert stack.hooks == {}
        assert stack.parameters == {}
        assert stack.sceptre_user_data == {}
        assert stack.template_path is None
        assert stack.template_handler_config == sentinel.template_handler_config
        assert stack.s3_details is None
        assert stack._template is None
        assert stack.protected is False
        assert stack.iam_role is None
        assert stack.role_arn is None
        assert stack.dependencies == []
        assert stack.tags == {}
        assert stack.notifications == []
        assert stack.on_failure is None
        assert stack.stack_group_config == {}

    def test_raises_exception_if_path_and_handler_configured(self):
        with pytest.raises(InvalidConfigFileError):
            Stack(
                name="stack_name", project_code="project_code",
                template_path="template_path", template_handler_config={"type": "file"},
                region="region"
            )

    def test_init__non_boolean_ignore_value__raises_invalid_config_file_error(self):
        with pytest.raises(InvalidConfigFileError):
            Stack(
                name='dev/stack/app', project_code=sentinel.project_code,
                template_handler_config=sentinel.template_handler_config,
                template_bucket_name=sentinel.template_bucket_name,
                template_key_prefix=sentinel.template_key_prefix,
                required_version=sentinel.required_version,
                region=sentinel.region, external_name=sentinel.external_name,
                ignore="true"
            )

    def test_init__non_boolean_obsolete_value__raises_invalid_config_file_error(self):
        with pytest.raises(InvalidConfigFileError):
            Stack(
                name='dev/stack/app', project_code=sentinel.project_code,
                template_handler_config=sentinel.template_handler_config,
                template_bucket_name=sentinel.template_bucket_name,
                template_key_prefix=sentinel.template_key_prefix,
                required_version=sentinel.required_version,
                region=sentinel.region, external_name=sentinel.external_name,
                obsolete="true"
            )

    def test_stack_repr(self):
        assert self.stack.__repr__() == \
            "sceptre.stack.Stack(" \
            "name='dev/app/stack', " \
            "project_code=sentinel.project_code, " \
            "template_path=sentinel.template_path, " \
            "template_handler_config=None, " \
            "region=sentinel.region, " \
            "template_bucket_name=sentinel.template_bucket_name, "\
            "template_key_prefix=sentinel.template_key_prefix, "\
            "required_version=sentinel.required_version, "\
            "iam_role=sentinel.iam_role, "\
            "iam_role_session_duration=sentinel.iam_role_session_duration, "\
            "profile=sentinel.profile, " \
            "sceptre_user_data=sentinel.sceptre_user_data, " \
            "parameters={'key1': 'val1'}, "\
            "hooks={}, "\
            "s3_details=None, " \
            "dependencies=sentinel.dependencies, "\
            "role_arn=sentinel.role_arn, "\
            "protected=False, "\
            "tags={'tag1': 'val1'}, "\
            "external_name=sentinel.external_name, " \
            "notifications=[sentinel.notification], " \
            "on_failure=sentinel.on_failure, " \
            "stack_timeout=sentinel.stack_timeout, " \
            "stack_group_config={}, " \
            "ignore=False, " \
            "obsolete=False" \
            ")"

    def test_configuration_manager__iam_role_raises_recursive_resolve__returns_connection_manager_with_no_role(self):
        class FakeResolver(Resolver):
            def resolve(self):
                return self.stack.iam_role

        self.stack.iam_role = FakeResolver()

        connection_manager = self.stack.connection_manager
        assert connection_manager.iam_role is None

    def test_configuration_manager__iam_role_returns_value_second_access__returns_value_on_second_access(self):

        class FakeResolver(Resolver):
            access_count = 0

            def resolve(self):
                if self.access_count == 0:
                    self.access_count += 1
                    return self.stack.iam_role
                else:
                    return 'role'

        self.stack.iam_role = FakeResolver()

        assert self.stack.connection_manager.iam_role is None
        assert self.stack.connection_manager.iam_role == 'role'

    def test_configuration_manager__iam_role_returns_value__returns_connection_manager_with_that_role(self):
        class FakeResolver(Resolver):
            def resolve(self):
                return 'role'

        self.stack.iam_role = FakeResolver()

        connection_manager = self.stack.connection_manager
        assert connection_manager.iam_role == 'role'


class TestStackSceptreUserData(object):
    def test_user_data_is_accessible(self):
        """
        .sceptre_user_data is a property. Let's make sure it accesses the right
        data.
        """
        stack = stack_factory(sceptre_user_data={'test_key': sentinel.test_value})
        assert stack.sceptre_user_data['test_key'] is sentinel.test_value

    def test_user_data_gets_resolved(self):
        class TestResolver(Resolver):
            def setup(self):
                pass

            def resolve(self):
                return sentinel.resolved_value

        stack = stack_factory(sceptre_user_data={'test_key': TestResolver()})
        assert stack.sceptre_user_data['test_key'] is sentinel.resolved_value

    def test_recursive_user_data_gets_resolved(self):
        """
        .sceptre_user_data can have resolvers that refer to .sceptre_user_data itself.
        Those must be instantiated before the attribute can be used.
        """
        class TestResolver(Resolver):
            def setup(self):
                pass

            def resolve(self):
                return self.stack.sceptre_user_data['primitive']

        stack = stack_factory()
        stack.sceptre_user_data = {
            'primitive': sentinel.primitive_value,
            'resolved': TestResolver(stack=stack),
        }
        assert stack.sceptre_user_data['resolved'] == sentinel.primitive_value



# -*- coding: utf-8 -*-

"""
sceptre.context

This module implements the SceptreContext class which holds details about the
paths used in a Sceptre project.
"""
from copy import deepcopy
from os import path

from sceptre.helpers import normalise_path


class SceptreContext(object):
    """
    SceptreContext is a place that holds data that is relevant to the
    project, including references to the project paths such as the path to your
    Sceptre project, templates path, config path, and the default names for
    your configuration files.

    :param project_path: Absolute path to the base sceptre project folder
    :type project_path: str

    :param command_path: The relative path to either StackGroup or Stack.
    :type command_path: str

    :param user_variables: Used to replace the value of anyvitem in a Config\
            file with a value defined by the CLI flag or in a YAML variable\
            file
    :type user_variables: dict

    :param options: The options specified in by the CLI command
    :type options: dict

    :param output_format: Specify the output format. Available formats:\
            [yaml, json]
    :type output_format: str

    :param no_colour: Specify whether colouring should be used in the CLI\
            output
    :type no_colour: bool

    :param full_scan: Specify whether folder scan the config files\
            True for scan all the config files and False for scan only in the command path
    :type full_scan: bool
    """

    def __init__(self, project_path, command_path,
                 user_variables=None, options=None, output_format=None,
                 no_colour=False, ignore_dependencies=False, full_scan=False):
        # project_path: absolute path to the base sceptre project folder
        # e.g. absolute_path/to/sceptre_directory
        self.project_path = normalise_path(project_path)

        # config_path: holds the project stack_groups
        # e.g {project_path}/config
        self.config_path = "config"  # user definable later in v2

        # command_path path to either stack group or stack
        # e.g. {project_path/config_path}/command_path
        self.command_path = normalise_path(command_path)

        self.normal_command_path = normalise_path(command_path)

        # config_file: stack group config. User definable later in v2
        # e.g. {project_path/config/command_path}/config_file
        self.config_file = "config.yaml"

        # templates_path: holds tempaltes. User definable later in v2
        # e.g. {project_path/}templates
        self.templates_path = "templates"

        self.user_variables = user_variables if user_variables else {}
        self.user_variables = user_variables\
            if user_variables is not None else {}
        self.options = options if options else {}
        self.output_format = output_format if output_format else ""
        self.no_colour = no_colour if no_colour is True else False
        self.ignore_dependencies = ignore_dependencies if ignore_dependencies is True else False
        self.full_scan = full_scan if full_scan is True else False

    def full_config_path(self):
        """
        Returns the config path in the format: ``project_path/config_path``.

        :returns: The absolute path to the config directory
        :rtype: str
        """
        return path.join(self.project_path, self.config_path)

    def full_command_path(self):
        """
        Returns the command path in the format:
        ``project_path/config_path/command_path``.

        :returns: The absolute path to the path that will be executed
        :rtype: str
        """
        return path.join(self.project_path, self.config_path,
                         self.command_path)

    def full_templates_path(self):
        """
        Returns the templates path in the format: project_path/templates_path.

        :returns: The absolute path to the  templates directory
        :rtype: str
        """
        return path.join(self.project_path, self.templates_path)

    def command_path_is_stack(self):
        """
        Returns True if the command path is a file.

        :returns: True if the command path is a file
        :rtype: bool
        """
        return path.isfile(
            path.join(
                self.project_path,
                self.config_path,
                self.command_path
            )
        )

    def clone(self) -> "SceptreContext":
        """Creates a new, deep clone of the context with all the same values."""
        new = type(self).__new__(type(self))
        new.__dict__.update(deepcopy(self.__dict__))
        return new



import abc
import logging
from functools import wraps

from sceptre.helpers import _call_func_on_values


class Hook(object):
    """
    Hook is an abstract base class that should be inherited by all hooks.

    :param argument: The argument of the hook.
    :type argument: str
    :param stack: The associated stack of the hook.
    :type stack: sceptre.stack.Stack
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, argument=None, stack=None):
        self.logger = logging.getLogger(__name__)
        self.argument = argument
        self.stack = stack

    def setup(self):
        """
        setup is a method that may be overwritten by inheriting classes. Allows
        hooks to run so initalisation steps when config is first read.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def run(self):
        """
        run is an abstract method which must be overwritten by all
        inheriting classes. Run should execute the logic of the hook.
        """
        pass  # pragma: no cover


class HookProperty(object):
    """
    This is a descriptor class used to store an attribute that may contain
    Hook objects. Used to setup Hooks when added as a attribute. Supports
    nested dictionary and lists.

    :param name: Attribute suffix used to store the property in the instance.
    :type name: str
    """

    def __init__(self, name):
        self.name = "_" + name
        self.logger = logging.getLogger(__name__)

    def __get__(self, instance, type):
        """
        Attribute getter for Hook containing data structure.

        :return: The attribute stored with the suffix ``name`` in the instance.
        :rtype: dict or list
        """
        return getattr(instance, self.name)

    def __set__(self, instance, value):
        """
        Attribute setter which adds a stack reference to any hooks in the
        data structure `value` and calls the setup method.

        """
        def setup(attr, key, value):
            value.stack = instance
            value.setup()

        _call_func_on_values(setup, value, Hook)
        setattr(instance, self.name, value)


def execute_hooks(hooks):
    """
    Searches through dictionary or list for Resolver objects and replaces
    them with the resolved value. Supports nested dictionaries and lists.
    Does not detect Resolver objects used as keys in dictionaries.

    :param attr: A complex data structure to search through.
    :type attr: dict or list
    :return: A complex data structure without Resolver objects.
    :rtype: dict or list
    """
    if isinstance(hooks, list):
        for hook in hooks:
            if isinstance(hook, Hook):
                hook.run()


def add_stack_hooks(func):
    """
    A function decorator to trigger the before and after hooks, relative
    to the decorated function's name.
    :param func: a function that operates on a stack
    :type func: function
    """
    @wraps(func)
    def decorated(self, *args, **kwargs):
        execute_hooks(self.stack.hooks.get("before_" + func.__name__))
        response = func(self, *args, **kwargs)
        execute_hooks(self.stack.hooks.get("after_" + func.__name__))

        return response

    return decorated



import click

from sceptre.context import SceptreContext
from sceptre.cli.helpers import catch_exceptions, confirmation
from sceptre.plan.plan import SceptrePlan


@click.command(name="execute")
@click.argument("path")
@click.argument("change-set-name")
@click.option(
    "-y", "--yes", is_flag=True, help="Assume yes to all questions."
)
@click.pass_context
@catch_exceptions
def execute_command(ctx, path, change_set_name, yes):
    """
    Executes a Change Set.
    \f

    :param path: Path to execute the command on.
    :type path: str
    :param change_set_name: Change Set to use.
    :type change_set_name: str
    :param yes: A flag to answer 'yes' too all CLI questions.
    :type yes: bool
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies")
    )

    plan = SceptrePlan(context)
    confirmation(
        plan.execute_change_set.__name__,
        yes,
        change_set=change_set_name,
        command_path=path
    )
    plan.execute_change_set(change_set_name)



# -*- coding: utf-8 -*-
import datetime
import json
from unittest.mock import patch, sentinel, Mock, call

import pytest
from botocore.exceptions import ClientError
from dateutil.tz import tzutc

from sceptre.exceptions import (
    CannotUpdateFailedStackError,
    ProtectedStackError,
    StackDoesNotExistError,
    UnknownStackChangeSetStatusError,
    UnknownStackStatusError
)
from sceptre.plan.actions import StackActions
from sceptre.stack import Stack
from sceptre.stack_status import StackChangeSetStatus, StackStatus
from sceptre.template import Template


class TestStackActions(object):

    def setup_method(self, test_method):
        self.patcher_connection_manager = patch(
            "sceptre.plan.actions.ConnectionManager"
        )
        self.mock_ConnectionManager = self.patcher_connection_manager.start()
        self.stack = Stack(
            name='prod/app/stack', project_code=sentinel.project_code,
            template_path=sentinel.template_path, region=sentinel.region,
            profile=sentinel.profile, parameters={"key1": "val1"}, sceptre_user_data=sentinel.sceptre_user_data,
            hooks={}, s3_details=None, dependencies=sentinel.dependencies,
            role_arn=sentinel.role_arn, protected=False,
            tags={"tag1": "val1"}, external_name=sentinel.external_name,
            notifications=[sentinel.notification],
            on_failure=sentinel.on_failure,
            stack_timeout=sentinel.stack_timeout,

        )
        self.actions = StackActions(self.stack)
        self.stack_group_config = {}
        self.template = Template(
            "fixtures/templates", self.stack.template_handler_config,
            self.stack.sceptre_user_data, self.stack_group_config,
            self.actions.connection_manager, self.stack.s3_details
        )
        self.template._body = json.dumps({
            'AWSTemplateFormatVersion': '2010-09-09',
            'Resources': {
                'Bucket': {
                    'Type': 'AWS::S3::Bucket',
                    'Properties': {}
                }
            }
        })
        self.stack._template = self.template

    def teardown_method(self, test_method):
        self.patcher_connection_manager.stop()

    @patch("sceptre.stack.Template")
    def test_template_loads_template(self, mock_Template):
        self.stack._template = None
        mock_Template.return_value = sentinel.template
        response = self.stack.template

        mock_Template.assert_called_once_with(
            name='prod/app/stack',
            handler_config={
                "type": "file",
                "path": sentinel.template_path
            },
            sceptre_user_data=sentinel.sceptre_user_data,
            stack_group_config={},
            connection_manager=self.stack.connection_manager,
            s3_details=None
        )
        assert response == sentinel.template

    def test_template_returns_template_if_it_exists(self):
        self.actions.stack._template = sentinel.template
        response = self.actions.stack._template
        assert response == sentinel.template

    def test_external_name_with_custom_stack_name(self):
        stack = Stack(
            name="stack_name", project_code="project_code",
            template_path="template_path", region="region",
            external_name="external_name"
        )

        assert stack.external_name == "external_name"

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    @patch("sceptre.plan.actions.StackActions._get_stack_timeout")
    def test_create_sends_correct_request(
            self, mock_get_stack_timeout, mock_wait_for_completion
    ):
        self.template._body = sentinel.template

        mock_get_stack_timeout.return_value = {
            "TimeoutInMinutes": sentinel.timeout
        }

        self.actions.create()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "TemplateBody": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM',
                                 'CAPABILITY_NAMED_IAM',
                                 'CAPABILITY_AUTO_EXPAND'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ],
                "OnFailure": sentinel.on_failure,
                "TimeoutInMinutes": sentinel.timeout
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    def test_create_sends_correct_request_no_notifications(
        self, mock_wait_for_completion
    ):
        self.actions.stack._template = Mock(spec=Template)
        self.actions.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }
        self.actions.stack.notifications = []

        self.actions.create()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM',
                                 'CAPABILITY_NAMED_IAM',
                                 'CAPABILITY_AUTO_EXPAND'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ],
                "OnFailure": sentinel.on_failure,
                "TimeoutInMinutes": sentinel.stack_timeout
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    def test_create_sends_correct_request_with_no_failure_no_timeout(
        self, mock_wait_for_completion
    ):
        self.template._body = sentinel.template
        self.actions.stack.on_failure = None
        self.actions.stack.stack_timeout = 0

        self.actions.create()

        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "TemplateBody": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM',
                                 'CAPABILITY_NAMED_IAM',
                                 'CAPABILITY_AUTO_EXPAND'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    def test_create_stack_already_exists(
            self, mock_wait_for_completion
    ):
        self.actions.stack._template = Mock(spec=Template)
        self.actions.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }
        mock_wait_for_completion.side_effect = ClientError(
            {
                "Error": {
                    "Code": "AlreadyExistsException",
                    "Message": "Stack already [{}] exists".format(self.actions.stack.name)
                }
            },
            sentinel.operation
        )
        response = self.actions.create()
        assert response == StackStatus.COMPLETE

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    def test_update_sends_correct_request(self, mock_wait_for_completion):
        self.actions.stack._template = Mock(spec=Template)
        self.actions.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }

        self.actions.update()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="update_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM',
                                 'CAPABILITY_NAMED_IAM',
                                 'CAPABILITY_AUTO_EXPAND'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )
        mock_wait_for_completion.assert_called_once_with(
            sentinel.stack_timeout
        )

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    def test_update_cancels_after_timeout(self, mock_wait_for_completion):
        self.actions.stack._template = Mock(spec=Template)
        self.actions.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }
        mock_wait_for_completion.return_value = StackStatus.IN_PROGRESS

        self.actions.update()
        calls = [
            call(
                service="cloudformation",
                command="update_stack",
                kwargs={
                    "StackName": sentinel.external_name,
                    "Template": sentinel.template,
                    "Parameters": [{
                        "ParameterKey": "key1",
                        "ParameterValue": "val1"
                    }],
                    "Capabilities": ['CAPABILITY_IAM',
                                     'CAPABILITY_NAMED_IAM',
                                     'CAPABILITY_AUTO_EXPAND'],
                    "RoleARN": sentinel.role_arn,
                    "NotificationARNs": [sentinel.notification],
                    "Tags": [
                        {"Key": "tag1", "Value": "val1"}
                    ]
                }),
            call(
                service="cloudformation",
                command="cancel_update_stack",
                kwargs={"StackName": sentinel.external_name})
        ]
        self.actions.connection_manager.call.assert_has_calls(calls)
        mock_wait_for_completion.assert_has_calls(
            [call(sentinel.stack_timeout), call()]
        )

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    def test_update_sends_correct_request_no_notification(
            self, mock_wait_for_completion
    ):
        self.actions.stack._template = Mock(spec=Template)
        self.actions.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }

        self.actions.stack.notifications = []
        self.actions.update()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="update_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM',
                                 'CAPABILITY_NAMED_IAM',
                                 'CAPABILITY_AUTO_EXPAND'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )
        mock_wait_for_completion.assert_called_once_with(
            sentinel.stack_timeout
        )

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    def test_update_with_complete_stack_with_no_updates_to_perform(
            self, mock_wait_for_completion
    ):
        self.actions.stack._template = Mock(spec=Template)
        self.actions.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }
        mock_wait_for_completion.side_effect = ClientError(
            {
                "Error": {
                    "Code": "NoUpdateToPerformError",
                    "Message": "No updates are to be performed."
                }
            },
            sentinel.operation
        )
        response = self.actions.update()
        assert response == StackStatus.COMPLETE

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    def test_cancel_update_sends_correct_request(
            self, mock_wait_for_completion
    ):
        self.actions.cancel_stack_update()
        self.actions.connection_manager.call.assert_called_once_with(
            service="cloudformation",
            command="cancel_update_stack",
            kwargs={"StackName": sentinel.external_name}
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("sceptre.plan.actions.StackActions.create")
    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_launch_with_stack_that_does_not_exist(
            self, mock_get_status, mock_create
    ):
        mock_get_status.side_effect = StackDoesNotExistError()
        mock_create.return_value = sentinel.launch_response
        response = self.actions.launch()
        mock_create.assert_called_once_with()
        assert response == sentinel.launch_response

    @patch("sceptre.plan.actions.StackActions.create")
    @patch("sceptre.plan.actions.StackActions.delete")
    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_launch_with_stack_that_failed_to_create(
            self, mock_get_status, mock_delete, mock_create
    ):
        mock_get_status.return_value = "CREATE_FAILED"
        mock_create.return_value = sentinel.launch_response
        response = self.actions.launch()
        mock_delete.assert_called_once_with()
        mock_create.assert_called_once_with()
        assert response == sentinel.launch_response

    @patch("sceptre.plan.actions.StackActions.update")
    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_launch_with_complete_stack_with_updates_to_perform(
            self, mock_get_status, mock_update
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_update.return_value = sentinel.launch_response
        response = self.actions.launch()
        mock_update.assert_called_once_with()
        assert response == sentinel.launch_response

    @patch("sceptre.plan.actions.StackActions.update")
    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_launch_with_complete_stack_with_no_updates_to_perform(
            self, mock_get_status, mock_update
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_update.return_value = StackStatus.COMPLETE
        response = self.actions.launch()
        mock_update.assert_called_once_with()
        assert response == StackStatus.COMPLETE

    @patch("sceptre.plan.actions.StackActions.update")
    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_launch_with_complete_stack_with_unknown_client_error(
            self, mock_get_status, mock_update
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_update.side_effect = ClientError(
            {
                "Error": {
                    "Code": "Boom!",
                    "Message": "Boom!"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.actions.launch()

    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_launch_with_in_progress_stack(self, mock_get_status):
        mock_get_status.return_value = "CREATE_IN_PROGRESS"
        response = self.actions.launch()
        assert response == StackStatus.IN_PROGRESS

    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_launch_with_failed_stack(self, mock_get_status):
        mock_get_status.return_value = "UPDATE_FAILED"
        with pytest.raises(CannotUpdateFailedStackError):
            response = self.actions.launch()
            assert response == StackStatus.FAILED

    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_launch_with_unknown_stack_status(self, mock_get_status):
        mock_get_status.return_value = "UNKNOWN_STATUS"
        with pytest.raises(UnknownStackStatusError):
            self.actions.launch()

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_delete_with_created_stack(
            self, mock_get_status, mock_wait_for_completion
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"

        self.actions.delete()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="delete_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "RoleARN": sentinel.role_arn
            }
        )

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_delete_when_wait_for_completion_raises_stack_does_not_exist_error(
            self, mock_get_status, mock_wait_for_completion
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_wait_for_completion.side_effect = StackDoesNotExistError()
        status = self.actions.delete()
        assert status == StackStatus.COMPLETE

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_delete_when_wait_for_completion_raises_non_existent_client_error(
            self, mock_get_status, mock_wait_for_completion
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_wait_for_completion.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Stack does not exist"
                }
            },
            sentinel.operation
        )
        status = self.actions.delete()
        assert status == StackStatus.COMPLETE

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_delete_when_wait_for_completion_raises_unexpected_client_error(
            self, mock_get_status, mock_wait_for_completion
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_wait_for_completion.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Boom"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.actions.delete()

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_delete_with_non_existent_stack(
            self, mock_get_status, mock_wait_for_completion
    ):
        mock_get_status.side_effect = StackDoesNotExistError()
        status = self.actions.delete()
        assert status == StackStatus.COMPLETE

    def test_describe_stack_sends_correct_request(self):
        self.actions.describe()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_stacks",
            kwargs={"StackName": sentinel.external_name}
        )

    def test_describe_events_sends_correct_request(self):
        self.actions.describe_events()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_stack_events",
            kwargs={"StackName": sentinel.external_name}
        )

    def test_describe_resources_sends_correct_request(self):
        self.actions.connection_manager.call.return_value = {
            "StackResources": [
                {
                    "LogicalResourceId": sentinel.logical_resource_id,
                    "PhysicalResourceId": sentinel.physical_resource_id,
                    "OtherParam": sentinel.other_param
                }
            ]
        }
        response = self.actions.describe_resources()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_stack_resources",
            kwargs={"StackName": sentinel.external_name}
        )
        assert response == {self.stack.name: [
            {
                "LogicalResourceId": sentinel.logical_resource_id,
                "PhysicalResourceId": sentinel.physical_resource_id
            }
        ]}

    @patch("sceptre.plan.actions.StackActions._describe")
    def test_describe_outputs_sends_correct_request(self, mock_describe):
        mock_describe.return_value = {
            "Stacks": [{
                "Outputs": sentinel.outputs
            }]
        }
        response = self.actions.describe_outputs()
        mock_describe.assert_called_once_with()
        assert response == {self.stack.name: sentinel.outputs}

    @patch("sceptre.plan.actions.StackActions._describe")
    def test_describe_outputs_handles_stack_with_no_outputs(
            self, mock_describe
    ):
        mock_describe.return_value = {
            "Stacks": [{}]
        }
        response = self.actions.describe_outputs()
        assert response == {self.stack.name: []}

    def test_continue_update_rollback_sends_correct_request(self):
        self.actions.continue_update_rollback()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="continue_update_rollback",
            kwargs={
                "StackName": sentinel.external_name,
                "RoleARN": sentinel.role_arn
            }
        )

    def test_set_stack_policy_sends_correct_request(self):
        self.actions.set_policy("tests/fixtures/stack_policies/unlock.json")
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="set_stack_policy",
            kwargs={
                "StackName": sentinel.external_name,
                "StackPolicyBody": """{
  "Statement" : [
    {
      "Effect" : "Allow",
      "Action" : "Update:*",
      "Principal": "*",
      "Resource" : "*"
    }
  ]
}
"""
            }
        )

    @patch("sceptre.plan.actions.json")
    def test_get_stack_policy_sends_correct_request(self, mock_Json):
        mock_Json.loads.return_value = '{}'
        mock_Json.dumps.return_value = '{}'
        response = self.actions.get_policy()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="get_stack_policy",
            kwargs={
                "StackName": sentinel.external_name
            }
        )

        assert response == {'prod/app/stack': '{}'}

    def test_create_change_set_sends_correct_request(self):
        self.template._body = sentinel.template

        self.actions.create_change_set(sentinel.change_set_name)
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_change_set",
            kwargs={
                "StackName": sentinel.external_name,
                "TemplateBody": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM',
                                 'CAPABILITY_NAMED_IAM',
                                 'CAPABILITY_AUTO_EXPAND'],
                "ChangeSetName": sentinel.change_set_name,
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )

    def test_create_change_set_sends_correct_request_no_notifications(self):
        self.actions.stack._template = Mock(spec=Template)
        self.actions.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }
        self.actions.stack.notifications = []

        self.actions.create_change_set(sentinel.change_set_name)
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_change_set",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM',
                                 'CAPABILITY_NAMED_IAM',
                                 'CAPABILITY_AUTO_EXPAND'],
                "ChangeSetName": sentinel.change_set_name,
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )

    def test_delete_change_set_sends_correct_request(self):
        self.actions.delete_change_set(sentinel.change_set_name)
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="delete_change_set",
            kwargs={
                "ChangeSetName": sentinel.change_set_name,
                "StackName": sentinel.external_name
            }
        )

    def test_describe_change_set_sends_correct_request(self):
        self.actions.describe_change_set(sentinel.change_set_name)
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_change_set",
            kwargs={
                "ChangeSetName": sentinel.change_set_name,
                "StackName": sentinel.external_name
            }
        )

    @patch("sceptre.plan.actions.StackActions._wait_for_completion")
    def test_execute_change_set_sends_correct_request(
        self, mock_wait_for_completion
    ):
        self.actions.execute_change_set(sentinel.change_set_name)
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="execute_change_set",
            kwargs={
                "ChangeSetName": sentinel.change_set_name,
                "StackName": sentinel.external_name
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    def test_execute_change_set__change_set_is_failed_for_no_changes__returns_0(self):
        def fake_describe(service, command, kwargs):
            assert (service, command) == ('cloudformation', 'describe_change_set')
            return {
                'Status': 'FAILED',
                'StatusReason': "The submitted information didn't contain changes",
            }
        self.actions.connection_manager.call.side_effect = fake_describe
        result = self.actions.execute_change_set(sentinel.change_set_name)
        assert result == 0

    def test_execute_change_set__change_set_is_failed_for_no_updates__returns_0(self):
        def fake_describe(service, command, kwargs):
            assert (service, command) == ('cloudformation', 'describe_change_set')
            return {
                'Status': 'FAILED',
                'StatusReason': "No updates are to be performed",
            }

        self.actions.connection_manager.call.side_effect = fake_describe
        result = self.actions.execute_change_set(sentinel.change_set_name)
        assert result == 0

    def test_list_change_sets_sends_correct_request(self):
        self.actions.list_change_sets()
        self.actions.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="list_change_sets",
            kwargs={"StackName": sentinel.external_name}
        )

    @patch("sceptre.plan.actions.StackActions._list_change_sets")
    def test_list_change_sets(
        self, mock_list_change_sets
    ):
        mock_list_change_sets_return_value = {"Summaries": []}
        expected_responses = []

        for num in ["1", "2"]:
            response = [{
                "ChangeSetId": "mychangesetid{num}",
                "StackId": "mystackid{num}"
            }]
            mock_list_change_sets_return_value["Summaries"].append(response)
            expected_responses.append(response)

        mock_list_change_sets.return_value = mock_list_change_sets_return_value

        response = self.actions.list_change_sets(url=False)
        assert response == {"prod/app/stack": expected_responses}

    @patch("sceptre.plan.actions.urllib.parse.urlencode")
    @patch("sceptre.plan.actions.StackActions._list_change_sets")
    def test_list_change_sets_url_mode(
        self, mock_list_change_sets, mock_urlencode
    ):
        mock_list_change_sets_return_value = {"Summaries": []}
        mock_urlencode_side_effect = []
        expected_urls = []

        for num in ["1", "2"]:
            mock_list_change_sets_return_value["Summaries"].append({
                "ChangeSetId": "mychangesetid{num}",
                "StackId": "mystackid{num}"
            })
            urlencoded = "stackId=mystackid{num}&changeSetId=mychangesetid{num}"
            mock_urlencode_side_effect.append(urlencoded)
            expected_urls.append(
                "https://sentinel.region.console.aws.amazon.com/cloudformation/home?"
                f"region=sentinel.region#/stacks/changesets/changes?{urlencoded}"
            )

        mock_list_change_sets.return_value = mock_list_change_sets_return_value
        mock_urlencode.side_effect = mock_urlencode_side_effect

        response = self.actions.list_change_sets(url=True)
        assert response == {"prod/app/stack": expected_urls}

    @pytest.mark.parametrize("url_mode", [True, False])
    @patch("sceptre.plan.actions.StackActions._list_change_sets")
    def test_list_change_sets_empty(
        self, mock_list_change_sets, url_mode
    ):
        mock_list_change_sets.return_value = {"Summaries": []}
        response = self.actions.list_change_sets(url=url_mode)
        assert response == {"prod/app/stack": []}

    @patch("sceptre.plan.actions.StackActions.set_policy")
    @patch("os.path.join")
    def test_lock_calls_set_stack_policy_with_policy(
            self, mock_join, mock_set_policy
    ):
        mock_join.return_value = "tests/fixtures/stack_policies/lock.json"
        self.actions.lock()
        mock_set_policy.assert_called_once_with(
            "tests/fixtures/stack_policies/lock.json"
        )

    @patch("sceptre.plan.actions.StackActions.set_policy")
    @patch("os.path.join")
    def test_unlock_calls_set_stack_policy_with_policy(
            self, mock_join, mock_set_policy
    ):
        mock_join.return_value = "tests/fixtures/stack_policies/unlock.json"
        self.actions.unlock()
        mock_set_policy.assert_called_once_with(
            "tests/fixtures/stack_policies/unlock.json"
        )

    def test_format_parameters_with_sting_values(self):
        parameters = {
            "key1": "value1",
            "key2": "value2",
            "key3": "value3"
        }
        formatted_parameters = self.actions._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1"},
            {"ParameterKey": "key2", "ParameterValue": "value2"},
            {"ParameterKey": "key3", "ParameterValue": "value3"}
        ]

    def test_format_parameters_with_none_values(self):
        parameters = {
            "key1": None,
            "key2": None,
            "key3": None
        }
        formatted_parameters = self.actions._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == []

    def test_format_parameters_with_none_and_string_values(self):
        parameters = {
            "key1": "value1",
            "key2": None,
            "key3": "value3"
        }
        formatted_parameters = self.actions._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1"},
            {"ParameterKey": "key3", "ParameterValue": "value3"}
        ]

    def test_format_parameters_with_list_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": ["value4", "value5", "value6"],
            "key3": ["value7", "value8", "value9"]
        }
        formatted_parameters = self.actions._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key2", "ParameterValue": "value4,value5,value6"},
            {"ParameterKey": "key3", "ParameterValue": "value7,value8,value9"}
        ]

    def test_format_parameters_with_none_and_list_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": None,
            "key3": ["value7", "value8", "value9"]
        }
        formatted_parameters = self.actions._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key3", "ParameterValue": "value7,value8,value9"}
        ]

    def test_format_parameters_with_list_and_string_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": "value4",
            "key3": ["value5", "value6", "value7"]
        }
        formatted_parameters = self.actions._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key2", "ParameterValue": "value4"},
            {"ParameterKey": "key3", "ParameterValue": "value5,value6,value7"}
        ]

    def test_format_parameters_with_none_list_and_string_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": "value4",
            "key3": None
        }
        formatted_parameters = self.actions._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key2", "ParameterValue": "value4"},
        ]

    @patch("sceptre.plan.actions.StackActions._describe")
    def test_get_status_with_created_stack(self, mock_describe):
        mock_describe.return_value = {
            "Stacks": [{"StackStatus": "CREATE_COMPLETE"}]
        }
        status = self.actions.get_status()
        assert status == "CREATE_COMPLETE"

    @patch("sceptre.plan.actions.StackActions._describe")
    def test_get_status_with_non_existent_stack(self, mock_describe):
        mock_describe.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Stack does not exist"
                }
            },
            sentinel.operation
        )
        assert self.actions.get_status() == "PENDING"

    @patch("sceptre.plan.actions.StackActions._describe")
    def test_get_status_with_unknown_clinet_error(self, mock_describe):
        mock_describe.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Boom!"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.actions.get_status()

    def test_get_role_arn_without_role(self):
        self.actions.stack.role_arn = None
        assert self.actions._get_role_arn() == {}

    def test_get_role_arn_with_role(self):
        assert self.actions._get_role_arn() == {"RoleARN": sentinel.role_arn}

    def test_protect_execution_without_protection(self):
        # Function should do nothing if protect == False
        self.actions._protect_execution()

    def test_protect_execution_without_explicit_protection(self):
        self.actions._protect_execution()

    def test_protect_execution_with_protection(self):
        self.actions.stack.protected = True
        with pytest.raises(ProtectedStackError):
            self.actions._protect_execution()

    @patch("sceptre.plan.actions.StackActions._log_new_events")
    @patch("sceptre.plan.actions.StackActions._get_status")
    @patch("sceptre.plan.actions.StackActions._get_simplified_status")
    def test_wait_for_completion_calls_log_new_events(
            self, mock_get_simplified_status, mock_get_status,
            mock_log_new_events
    ):
        mock_get_simplified_status.return_value = StackStatus.COMPLETE

        self.actions._wait_for_completion()
        mock_log_new_events.assert_called_once_with()

    @pytest.mark.parametrize("test_input,expected", [
        ("ROLLBACK_COMPLETE", StackStatus.FAILED),
        ("STACK_COMPLETE", StackStatus.COMPLETE),
        ("STACK_IN_PROGRESS", StackStatus.IN_PROGRESS),
        ("STACK_FAILED", StackStatus.FAILED)
    ])
    def test_get_simplified_status_with_known_stack_statuses(
            self, test_input, expected
    ):
        response = self.actions._get_simplified_status(test_input)
        assert response == expected

    def test_get_simplified_status_with_stack_in_unknown_state(self):
        with pytest.raises(UnknownStackStatusError):
            self.actions._get_simplified_status("UNKOWN_STATUS")

    @patch("sceptre.plan.actions.StackActions.describe_events")
    def test_log_new_events_calls_describe_events(self, mock_describe_events):
        mock_describe_events.return_value = {
            "StackEvents": []
        }
        self.actions._log_new_events()
        self.actions.describe_events.assert_called_once_with()

    @patch("sceptre.plan.actions.StackActions.describe_events")
    def test_log_new_events_prints_correct_event(self, mock_describe_events):
        self.actions.stack.name = "stack-name"
        mock_describe_events.return_value = {
            "StackEvents": [
                {
                    "Timestamp": datetime.datetime(
                        2016, 3, 15, 14, 2, 0, 0, tzinfo=tzutc()
                    ),
                    "LogicalResourceId": "id-2",
                    "ResourceType": "type-2",
                    "ResourceStatus": "resource-status"
                },
                {
                    "Timestamp": datetime.datetime(
                        2016, 3, 15, 14, 1, 0, 0, tzinfo=tzutc()
                    ),
                    "LogicalResourceId": "id-1",
                    "ResourceType": "type-1",
                    "ResourceStatus": "resource",
                    "ResourceStatusReason": "User Initiated"
                }
            ]
        }
        self.actions.most_recent_event_datetime = (
            datetime.datetime(2016, 3, 15, 14, 0, 0, 0, tzinfo=tzutc())
        )
        self.actions._log_new_events()

    @patch("sceptre.plan.actions.StackActions._get_cs_status")
    def test_wait_for_cs_completion_calls_get_cs_status(
        self, mock_get_cs_status
    ):
        mock_get_cs_status.side_effect = [
            StackChangeSetStatus.PENDING, StackChangeSetStatus.READY
        ]

        self.actions.wait_for_cs_completion(sentinel.change_set_name)
        mock_get_cs_status.assert_called_with(sentinel.change_set_name)

    @patch("sceptre.plan.actions.StackActions.describe_change_set")
    def test_get_cs_status_handles_all_statuses(
        self, mock_describe_change_set
    ):
        scss = StackChangeSetStatus
        return_values = {                                                                                                     # NOQA
                 "Status":    ('CREATE_PENDING', 'CREATE_IN_PROGRESS', 'CREATE_COMPLETE', 'DELETE_COMPLETE', 'FAILED'),       # NOQA
        "ExecutionStatus": {                                                                                                  # NOQA
        'UNAVAILABLE':         (scss.PENDING,     scss.PENDING,         scss.PENDING,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'AVAILABLE':           (scss.PENDING,     scss.PENDING,         scss.READY,        scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'EXECUTE_IN_PROGRESS': (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'EXECUTE_COMPLETE':    (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'EXECUTE_FAILED':      (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'OBSOLETE':            (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        }                                                                                                                     # NOQA
        }                                                                                                                     # NOQA

        for i, status in enumerate(return_values['Status']):
            for exec_status, returns in \
                    return_values['ExecutionStatus'].items():
                mock_describe_change_set.return_value = {
                    "Status": status,
                    "ExecutionStatus": exec_status
                }
                response = self.actions._get_cs_status(
                    sentinel.change_set_name
                )
                assert response == returns[i]

        for status in return_values['Status']:
            mock_describe_change_set.return_value = {
                "Status": status,
                "ExecutionStatus": 'UNKOWN_STATUS'
            }
            with pytest.raises(UnknownStackChangeSetStatusError):
                self.actions._get_cs_status(sentinel.change_set_name)

        for exec_status in return_values['ExecutionStatus'].keys():
            mock_describe_change_set.return_value = {
                "Status": 'UNKOWN_STATUS',
                "ExecutionStatus": exec_status
            }
            with pytest.raises(UnknownStackChangeSetStatusError):
                self.actions._get_cs_status(sentinel.change_set_name)

        mock_describe_change_set.return_value = {
            "Status": 'UNKOWN_STATUS',
            "ExecutionStatus": 'UNKOWN_STATUS',
        }
        with pytest.raises(UnknownStackChangeSetStatusError):
            self.actions._get_cs_status(sentinel.change_set_name)

    @patch("sceptre.plan.actions.StackActions.describe_change_set")
    def test_get_cs_status_raises_unexpected_exceptions(
        self, mock_describe_change_set
    ):
        mock_describe_change_set.side_effect = ClientError(
            {
                "Error": {
                    "Code": "ChangeSetNotFound",
                    "Message": "ChangeSet [*] does not exist"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.actions._get_cs_status(sentinel.change_set_name)

    def test_fetch_remote_template__cloudformation_returns_validation_error__returns_none(self):
        self.actions.connection_manager.call.side_effect = ClientError(
            {
                "Error": {
                    "Code": "ValidationError",
                    "Message": "An error occurred (ValidationError) "
                               "when calling the GetTemplate operation: "
                               "Stack with id foo does not exist"
                }
            },
            sentinel.operation
        )

        result = self.actions.fetch_remote_template()
        assert result is None

    def test_fetch_remote_template__calls_cloudformation_get_template(self):
        self.actions.connection_manager.call.return_value = {'TemplateBody': ''}
        self.actions.fetch_remote_template()

        self.actions.connection_manager.call.assert_called_with(
            service='cloudformation',
            command='get_template',
            kwargs={
                'StackName': self.stack.external_name,
                'TemplateStage': 'Original'
            }
        )

    def test_fetch_remote_template__dict_template__returns_json(self):
        template_body = {
            'AWSTemplateFormatVersion': '2010-09-09',
            'Resources': {}
        }
        self.actions.connection_manager.call.return_value = {
            'TemplateBody': template_body
        }
        expected = json.dumps(template_body, indent=4)

        result = self.actions.fetch_remote_template()
        assert result == expected

    def test_fetch_remote_template__cloudformation_returns_string_template__returns_that_string(self):
        template_body = "This is my template"
        self.actions.connection_manager.call.return_value = {
            'TemplateBody': template_body
        }
        result = self.actions.fetch_remote_template()
        assert result == template_body

    def test_fetch_remote_template_summary__calls_cloudformation_get_template_summary(self):
        self.actions.fetch_remote_template_summary()

        self.actions.connection_manager.call.assert_called_with(
            service='cloudformation',
            command='get_template_summary',
            kwargs={
                'StackName': self.stack.external_name,
            }
        )

    def test_fetch_remote_template_summary__returns_response_from_cloudformation(self):
        def get_template_summary(service, command, kwargs):
            assert (service, command) == ('cloudformation', 'get_template_summary')
            return {'template': 'summary'}

        self.actions.connection_manager.call.side_effect = get_template_summary
        result = self.actions.fetch_remote_template_summary()
        assert result == {'template': 'summary'}

    def test_fetch_local_template_summary__calls_cloudformation_get_template_summary(self):
        self.actions.fetch_local_template_summary()

        self.actions.connection_manager.call.assert_called_with(
            service='cloudformation',
            command='get_template_summary',
            kwargs={
                'TemplateBody': self.stack.template.body,
            }
        )

    def test_fetch_local_template_summary__returns_response_from_cloudformation(self):
        def get_template_summary(service, command, kwargs):
            assert (service, command) == ('cloudformation', 'get_template_summary')
            return {'template': 'summary'}

        self.actions.connection_manager.call.side_effect = get_template_summary
        result = self.actions.fetch_local_template_summary()
        assert result == {'template': 'summary'}

    def test_fetch_local_template_summary__cloudformation_returns_validation_error_invalid_stack__raises_it(self):
        self.actions.connection_manager.call.side_effect = ClientError(
            {
                "Error": {
                    "Code": "ValidationError",
                    "Message": "Template format error: Resource name {Invalid::Resource} is "
                               "non alphanumeric.'"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.actions.fetch_local_template_summary()

    def test_fetch_remote_template_summary__cloudformation_returns_validation_error_for_no_stack__returns_none(self):
        self.actions.connection_manager.call.side_effect = ClientError(
            {
                "Error": {
                    "Code": "ValidationError",
                    "Message": "An error occurred (ValidationError) "
                               "when calling the GetTemplate operation: "
                               "Stack with id foo does not exist"
                }
            },
            sentinel.operation
        )
        result = self.actions.fetch_remote_template_summary()
        assert result is None

    def test_diff__invokes_diff_method_on_injected_differ_with_self(self):
        differ = Mock()
        self.actions.diff(differ)
        differ.diff.assert_called_with(self.actions)

    def test_diff__returns_result_of_injected_differs_diff_method(self):
        differ = Mock()
        result = self.actions.diff(differ)
        assert result == differ.diff.return_value

    @patch("sceptre.plan.actions.StackActions._describe_stack_drift_detection_status")
    @patch("sceptre.plan.actions.StackActions._detect_stack_drift")
    @patch("time.sleep")
    def test_drift_detect(
        self,
        mock_sleep,
        mock_detect_stack_drift,
        mock_describe_stack_drift_detection_status
    ):
        mock_sleep.return_value = None

        mock_detect_stack_drift.return_value = {
            "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62"
        }

        first_response = {
            "StackId": "fake-stack-id",
            "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62",
            "DetectionStatus": "DETECTION_IN_PROGRESS",
            "StackDriftStatus": "NOT_CHECKED",
            "DetectionStatusReason": "User Initiated"
        }

        final_response = {
            "StackId": "fake-stack-id",
            "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62",
            "StackDriftStatus": "IN_SYNC",
            "DetectionStatus": "DETECTION_COMPLETE",
            "DriftedStackResourceCount": 0
        }

        mock_describe_stack_drift_detection_status.side_effect = [
            first_response, final_response
        ]

        response = self.actions.drift_detect()
        assert response == final_response

    @pytest.mark.parametrize("detection_status", ["DETECTION_COMPLETE", "DETECTION_FAILED"])
    @patch("sceptre.plan.actions.StackActions._describe_stack_resource_drifts")
    @patch("sceptre.plan.actions.StackActions._describe_stack_drift_detection_status")
    @patch("sceptre.plan.actions.StackActions._detect_stack_drift")
    @patch("time.sleep")
    def test_drift_show(
        self,
        mock_sleep,
        mock_detect_stack_drift,
        mock_describe_stack_drift_detection_status,
        mock_describe_stack_resource_drifts,
        detection_status
    ):
        mock_sleep.return_value = None

        mock_detect_stack_drift.return_value = {
            "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62"
        }
        mock_describe_stack_drift_detection_status.side_effect = [
            {
                "StackId": "fake-stack-id",
                "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62",
                "DetectionStatus": "DETECTION_IN_PROGRESS",
                "StackDriftStatus": "FOO",
                "DetectionStatusReason": "User Initiated"
            },
            {
                "StackId": "fake-stack-id",
                "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62",
                "StackDriftStatus": "FOO",
                "DetectionStatus": detection_status,
                "DriftedStackResourceCount": 0
            }
        ]

        expected_drifts = {
            "StackResourceDrifts": [
                {
                    "StackId": "fake-stack-id",
                    "LogicalResourceId": "VPC",
                    "PhysicalResourceId": "vpc-028c655dea7c65227",
                    "ResourceType": "AWS::EC2::VPC",
                    "ExpectedProperties": '{"foo":"bar"}',
                    "ActualProperties": '{"foo":"bar"}',
                    "PropertyDifferences": [],
                    "StackResourceDriftStatus": detection_status
                }
            ]
        }

        mock_describe_stack_resource_drifts.return_value = expected_drifts
        expected_response = (detection_status, expected_drifts)

        response = self.actions.drift_show(drifted=False)

        assert response == expected_response

    @patch("sceptre.plan.actions.StackActions._describe_stack_resource_drifts")
    @patch("sceptre.plan.actions.StackActions._describe_stack_drift_detection_status")
    @patch("sceptre.plan.actions.StackActions._detect_stack_drift")
    @patch("time.sleep")
    def test_drift_show_drift_only(
        self,
        mock_sleep,
        mock_detect_stack_drift,
        mock_describe_stack_drift_detection_status,
        mock_describe_stack_resource_drifts
    ):
        mock_sleep.return_value = None

        mock_detect_stack_drift.return_value = {
            "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62"
        }
        mock_describe_stack_drift_detection_status.return_value = {
            "StackId": "fake-stack-id",
            "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62",
            "StackDriftStatus": "DRIFTED",
            "DetectionStatus": "DETECTION_COMPLETE",
            "DriftedStackResourceCount": 0
        }

        input_drifts = {
            "StackResourceDrifts": [
                {
                    "LogicalResourceId": "ServerLoadBalancer",
                    "PhysicalResourceId": "bi-tablea-ServerLo-1E133TWLWYLON",
                    "ResourceType": "AWS::ElasticLoadBalancing::LoadBalancer",
                    "StackId": "fake-stack-id",
                    "StackResourceDriftStatus": "IN_SYNC",
                },
                {
                    "LogicalResourceId": "TableauServer",
                    "PhysicalResourceId": "i-08c16bc1c5e2cd185",
                    "ResourceType": "AWS::EC2::Instance",
                    "StackId": "fake-stack-id",
                    "StackResourceDriftStatus": "DELETED",
                }
            ]
        }
        mock_describe_stack_resource_drifts.return_value = input_drifts

        expected_response = ("DETECTION_COMPLETE", {
            "StackResourceDrifts": [input_drifts["StackResourceDrifts"][1]]
        })

        response = self.actions.drift_show(drifted=True)

        assert response == expected_response

    @patch("sceptre.plan.actions.StackActions._get_status")
    def test_drift_show_with_stack_that_does_not_exist(self, mock_get_status):
        mock_get_status.side_effect = StackDoesNotExistError()
        response = self.actions.drift_show(drifted=False)
        assert response == (
            'STACK_DOES_NOT_EXIST', {
                'StackResourceDriftStatus': 'STACK_DOES_NOT_EXIST'
            })

    @patch("sceptre.plan.actions.StackActions._describe_stack_resource_drifts")
    @patch("sceptre.plan.actions.StackActions._describe_stack_drift_detection_status")
    @patch("sceptre.plan.actions.StackActions._detect_stack_drift")
    @patch("time.sleep")
    def test_drift_show_times_out(
        self,
        mock_sleep,
        mock_detect_stack_drift,
        mock_describe_stack_drift_detection_status,
        mock_describe_stack_resource_drifts,
    ):
        mock_sleep.return_value = None

        mock_detect_stack_drift.return_value = {
            "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62"
        }

        response = {
            "StackId": "fake-stack-id",
            "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62",
            "DetectionStatus": "DETECTION_IN_PROGRESS",
            "StackDriftStatus": "FOO",
            "DetectionStatusReason": "User Initiated"
        }

        side_effect = []
        for _ in range(0, 60):
            side_effect.append(response)
        mock_describe_stack_drift_detection_status.side_effect = side_effect

        expected_drifts = {
            "StackResourceDrifts": [
                {
                    "StackId": "fake-stack-id",
                    "LogicalResourceId": "VPC",
                    "PhysicalResourceId": "vpc-028c655dea7c65227",
                    "ResourceType": "AWS::EC2::VPC",
                    "ExpectedProperties": '{"foo":"bar"}',
                    "ActualProperties": '{"foo":"bar"}',
                    "PropertyDifferences": [],
                    "StackResourceDriftStatus": "DETECTION_IN_PROGRESS"
                }
            ]
        }

        mock_describe_stack_resource_drifts.return_value = expected_drifts
        expected_response = ("TIMED_OUT", {"StackResourceDriftStatus": "TIMED_OUT"})

        response = self.actions.drift_show(drifted=False)

        assert response == expected_response



import pytest

from sceptre.resolvers import are_placeholders_enabled, Resolver
from sceptre.resolvers.placeholders import (
    use_resolver_placeholders_on_error,
    PlaceholderType,
    create_placeholder_value
)


class TestPlaceholders:

    def test_are_placeholders_enabled__returns_false(self):
        assert are_placeholders_enabled() is False

    def test_are_placeholders_enabled__in_placeholder_context__returns_true(self):
        with use_resolver_placeholders_on_error():
            assert are_placeholders_enabled() is True

    def test_are_placeholders_enabled__out_of_placeholder_context__returns_false(self):
        with use_resolver_placeholders_on_error():
            pass

        assert are_placeholders_enabled() is False

    def test_are_placeholders_enabled__error_in_placeholder_context__returns_false(self):
        with pytest.raises(ValueError), use_resolver_placeholders_on_error():
            raise ValueError()

        assert are_placeholders_enabled() is False

    @pytest.mark.parametrize(
        'placeholder_type,argument,expected',
        [
            pytest.param(PlaceholderType.explicit, None, '{ !MyResolver }', id='explicit no argument'),
            pytest.param(
                PlaceholderType.explicit,
                'argument',
                '{ !MyResolver(argument) }',
                id='explicit string argument'
            ),
            pytest.param(
                PlaceholderType.explicit,
                {'key': 'value'},
                "{ !MyResolver({'key': 'value'}) }",
                id='explicit dict argument'
            ),
            pytest.param(PlaceholderType.alphanum, None, 'MyResolver', id='alphanum no argument'),
            pytest.param(PlaceholderType.alphanum, 'argument', 'MyResolverargument', id='alphanum string argument'),
            pytest.param(PlaceholderType.alphanum, {'key': 'value'}, 'MyResolverkeyvalue', id='alphanum dict argument'),
            pytest.param(PlaceholderType.none, 'something', None)
        ]
    )
    def test_create_placeholder_value(self, placeholder_type, argument, expected):
        class MyResolver(Resolver):
            def resolve(self):
                pass

        resolver = MyResolver(argument)

        with use_resolver_placeholders_on_error():
            result = create_placeholder_value(resolver, placeholder_type)

        assert result == expected



from contextlib import contextmanager
from enum import Enum
from threading import Lock
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from sceptre import resolvers

# This is a toggle used for globally enabling placeholder values out of resolvers when they error
# while resolving. This is important when performing actions on stacks like validation or generation
# when their dependencies have not been deployed yet and those dependencies are expressed in stack
# resolvers that are used in those actions, especially sceptre_user_data.
_RESOLVE_PLACEHOLDER_ON_ERROR = False


class PlaceholderType(Enum):
    explicit = 1  # Looks like "{ !MyClass(argument) }"
    alphanum = 2  # Looks like MyClassargument
    none = 3  # Resolves to None


_placeholder_lock = Lock()


@contextmanager
def use_resolver_placeholders_on_error():
    """A context manager that toggles on placeholders for resolvers that error out. This should NOT
    be used while creating/launching stacks, but it is often required when validating or diffing
    stacks whose dependencies haven't yet been deployed and that reference those dependencies with
    resolvers, especially in the sceptre_user_data.
    """
    global _RESOLVE_PLACEHOLDER_ON_ERROR

    try:
        with _placeholder_lock:
            _RESOLVE_PLACEHOLDER_ON_ERROR = True
        yield
    finally:
        with _placeholder_lock:
            _RESOLVE_PLACEHOLDER_ON_ERROR = False


def are_placeholders_enabled() -> bool:
    """Indicates whether placeholders have been globally enabled or not."""
    with _placeholder_lock:
        return _RESOLVE_PLACEHOLDER_ON_ERROR


def create_placeholder_value(resolver: 'resolvers.Resolver', placeholder_type: PlaceholderType) -> Any:
    placeholder_func = _placeholders[placeholder_type]
    return placeholder_func(resolver)


def _create_explicit_resolver_placeholder(resolver: 'resolvers.Resolver') -> str:
    """Creates a placeholder value to be substituted for the resolved value when placeholders are
    allowed and the value cannot be resolved.

    The placeholder will look like one of:
      * { !ClassName } -> used when there is no argument
      * { !ClassName(argument) } -> used when there is a string argument
      * { !ClassName({'key': 'value'}) } -> used when there is a dict argument

    :param resolver: The resolver to create a placeholder for
    :return: The placeholder value
    """
    base = f'!{type(resolver).__name__}'
    suffix = f'({resolver.argument})' if resolver.argument is not None else ''
    # double-braces in an f-string is just an escaped single brace
    return f'{{ {base}{suffix} }}'


def _create_alphanumeric_placeholder(resolver: 'resolvers.Resolver') -> str:
    """Creates a placeholder value that is only composed of alphanumeric characters. This is more
    useful when performing operations that send a template to CloudFormation, which will have stricter
    requirements for values in templates.

    Values from this function will not be as readable, but they are more likely to be valid when passed
    to a template.

    The placeholder will look like one of:
    * ClassName -> used when there is no argument
    * ClassNameargument -> used when there is a string argument
    * ClassNamekeyvalue -> used when there is a dict argument

    :param resolver: The resolver to create a placeholder for
    :return: The placeholder value
    """
    explicit_placeholder = _create_explicit_resolver_placeholder(resolver)
    alphanum_placeholder = ''.join(c for c in explicit_placeholder if c.isalnum())
    return alphanum_placeholder


_placeholders = {
    PlaceholderType.explicit: _create_explicit_resolver_placeholder,
    PlaceholderType.alphanum: _create_alphanumeric_placeholder,
    PlaceholderType.none: lambda resolver: None
}



# -*- coding: utf-8 -*-
from unittest.mock import MagicMock

from sceptre.hooks import Hook, HookProperty, add_stack_hooks, execute_hooks


class MockHook(Hook):
    def __init__(self, *args, **kwargs):
        super(MockHook, self).__init__(*args, **kwargs)

    def run(self):
        pass


class TestHooksFunctions(object):
    def setup_method(self, test_method):
        pass

    def test_add_stack_hooks(self):
        mock_hook_before = MagicMock(spec=Hook)
        mock_hook_after = MagicMock(spec=Hook)
        mock_object = MagicMock()

        mock_object.stack.hooks = {
            'before_mock_function': [mock_hook_before],
            'after_mock_function': [mock_hook_after]
        }

        def mock_function(self):
            assert mock_hook_before.run.call_count == 1
            assert mock_hook_after.run.call_count == 0

        mock_object.mock_function = mock_function
        mock_object.mock_function.__name__ = 'mock_function'

        add_stack_hooks(mock_object.mock_function)(mock_object)

        assert mock_hook_before.run.call_count == 1
        assert mock_hook_after.run.call_count == 1

    def test_execute_hooks_with_not_a_list(self):
        execute_hooks(None)

    def test_execute_hooks_with_empty_list(self):
        execute_hooks([])

    def test_execute_hooks_with_list_with_non_hook_objects(self):
        execute_hooks([None, "string", 1, True])

    def test_execute_hooks_with_single_hook(self):
        hook = MagicMock(spec=Hook)
        execute_hooks([hook])
        hook.run.called_once_with()

    def test_execute_hooks_with_multiple_hook(self):
        hook_1 = MagicMock(spec=Hook)
        hook_2 = MagicMock(spec=Hook)
        execute_hooks([hook_1, hook_2])
        hook_1.run.called_once_with()
        hook_2.run.called_once_with()


class TestHook(object):
    def setup_method(self, test_method):
        self.hook = MockHook()

    def test_hook_inheritance(self):
        assert isinstance(self.hook, Hook)


class MockClass(object):
    hook_property = HookProperty("hook_property")
    config = MagicMock()


class TestHookPropertyDescriptor(object):

    def setup_method(self, test_method):
        self.mock_object = MockClass()

    def test_setting_hook_property(self):
        mock_hook = MagicMock(spec=MockHook)

        self.mock_object.hook_property = [mock_hook]
        assert self.mock_object._hook_property == [mock_hook]

    def test_getting_hook_property(self):
        self.mock_object._hook_property = self.mock_object
        assert self.mock_object.hook_property == self.mock_object



# -*- coding: utf-8 -*-
import json
import io
import pytest

from unittest.mock import MagicMock
from sceptre.connection_manager import ConnectionManager
from sceptre.exceptions import SceptreException, UnsupportedTemplateFileTypeError
from sceptre.template_handlers.s3 import S3
from unittest.mock import patch


class TestS3(object):

    def test_get_template(self):
        connection_manager = MagicMock(spec=ConnectionManager)
        connection_manager.call.return_value = {
            "Body": io.BytesIO(b"Stuff is working")
        }
        template_handler = S3(
            name="s3_handler",
            arguments={"path": "bucket/folder/file.yaml"},
            connection_manager=connection_manager
        )
        result = template_handler.handle()

        connection_manager.call.assert_called_once_with(
            service="s3",
            command="get_object",
            kwargs={
                "Bucket": "bucket",
                "Key": "folder/file.yaml"
            }
        )
        assert result == b"Stuff is working"

    def test_template_handler(self):
        connection_manager = MagicMock(spec=ConnectionManager)
        connection_manager.call.return_value = {
            "Body": io.BytesIO(b"Stuff is working")
        }
        template_handler = S3(
            name="vpc",
            arguments={"path": "my-fancy-bucket/account/vpc.yaml"},
            connection_manager=connection_manager
        )
        result = template_handler.handle()

        connection_manager.call.assert_called_once_with(
            service="s3",
            command="get_object",
            kwargs={
                "Bucket": "my-fancy-bucket",
                "Key": "account/vpc.yaml"
            }
        )
        assert result == b"Stuff is working"

    def test_invalid_response_reraises_exception(self):
        connection_manager = MagicMock(spec=ConnectionManager)
        connection_manager.call.side_effect = SceptreException("BOOM!")

        template_handler = S3(
            name="vpc",
            arguments={"path": "my-fancy-bucket/account/vpc.yaml"},
            connection_manager=connection_manager
        )

        with pytest.raises(SceptreException) as e:
            template_handler.handle()

        assert str(e.value) == "BOOM!"

    def test_handler_unsupported_type(self):
        s3_handler = S3("s3_handler", {'path': 'bucket/folder/file.unsupported'})
        with pytest.raises(UnsupportedTemplateFileTypeError):
            s3_handler.handle()

    @pytest.mark.parametrize("path", [
        ("bucket/folder/file.json"),
        ("bucket/folder/file.yaml"),
        ("bucket/folder/file.template")
    ])
    @patch('sceptre.template_handlers.s3.S3._get_template')
    def test_handler_raw_template(self, mock_get_template, path):
        mock_get_template.return_value = {}
        s3_handler = S3("s3_handler", {'path': path})
        s3_handler.handle()
        assert mock_get_template.call_count == 1

    @patch('sceptre.template_handlers.helper.render_jinja_template')
    @patch('sceptre.template_handlers.s3.S3._get_template')
    def test_handler_jinja_template(slef, mock_get_template, mock_render_jinja_template):
        mock_get_template_response = {
            "Description": "test template",
            "AWSTemplateFormatVersion": "2010-09-09",
            "Resources": {
                "touchNothing": {
                    "Type": "AWS::CloudFormation::WaitConditionHandle"
                }
            }
        }
        mock_get_template.return_value = json.dumps(mock_get_template_response).encode('utf-8')
        s3_handler = S3("s3_handler", {'path': 'bucket/folder/file.j2'})
        s3_handler.handle()
        assert mock_render_jinja_template.call_count == 1

    @patch('sceptre.template_handlers.helper.call_sceptre_handler')
    @patch('sceptre.template_handlers.s3.S3._get_template')
    def test_handler_python_template(self, mock_get_template, mock_call_sceptre_handler):
        mock_get_template_response = {
            "Description": "test template",
            "AWSTemplateFormatVersion": "2010-09-09",
            "Resources": {
                "touchNothing": {
                    "Type": "AWS::CloudFormation::WaitConditionHandle"
                }
            }
        }
        mock_get_template.return_value = json.dumps(mock_get_template_response).encode('utf-8')
        s3_handler = S3("s3_handler", {'path': 'bucket/folder/file.py'})
        s3_handler.handle()
        assert mock_call_sceptre_handler.call_count == 1



import logging
import click

from sceptre.context import SceptreContext
from sceptre.cli.helpers import (
    catch_exceptions,
    write
)
from sceptre.plan.plan import SceptrePlan

logger = logging.getLogger(__name__)


@click.group(name="dump")
def dump_group():
    """
    Commands for dumping attributes of stacks.
    """
    pass


@dump_group.command(name="config")
@click.argument("path")
@click.pass_context
@catch_exceptions
def dump_config(ctx, path):
    """
    Dump the rendered (post-Jinja) Stack Configs.
    \f

    :param path: Path to execute the command on or path to stack group
    """
    context = SceptreContext(
        command_path=path,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        output_format=ctx.obj.get("output_format"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies")
    )
    plan = SceptrePlan(context)
    responses = plan.dump_config()

    output = []
    for stack, config in responses.items():
        if config is None:
            logger.warning(f"{stack.external_name} does not exist")
        else:
            output.append({stack.external_name: config})

    output_format = "json" if context.output_format == "json" else "yaml"

    if len(output) == 1:
        write(output[0][stack.external_name], output_format)
    else:
        for config in output:
            write(config, output_format)
